# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
from karbor.context import RequestContext
from karbor.resource import Resource
from karbor.services.protection.protectable_plugins.server \
    import ServerProtectablePlugin

from karbor.tests import base
from keystoneauth1 import session as keystone_session
import mock
from novaclient.v2 import servers
from oslo_config import cfg


class ServerProtectablePluginTest(base.TestCase):
    def setUp(self):
        super(ServerProtectablePluginTest, self).setUp()
        service_catalog = [
            {'type': 'compute',
             'endpoints': [{'publicURL': 'http://127.0.0.1:8774/v2.1/abcd'}],
             },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_endpoint(self, mock_generate_session):
        cfg.CONF.set_default('nova_endpoint',
                             'http://127.0.0.1:8774/v2.1',
                             'nova_client')
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual('compute',
                         plugin._client(self._context).client.service_type)
        self.assertEqual(
            'http://127.0.0.1:8774/v2.1/abcd',
            plugin._client(self._context).client.endpoint_override)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_catalog(self, mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual('compute',
                         plugin._client(self._context).client.service_type)
        self.assertEqual(
            'http://127.0.0.1:8774/v2.1/abcd',
            plugin._client(self._context).client.endpoint_override)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_get_resource_type(self, mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual("OS::Nova::Server", plugin.get_resource_type())

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_get_parent_resource_types(self, mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual(("OS::Keystone::Project", ),
                         plugin.get_parent_resource_types())

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    @mock.patch.object(servers.ServerManager, 'list')
    def test_list_resources(self, mock_server_list, mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)

        server_info = collections.namedtuple('server_info', ['id', 'name',
                                                             'status'])
        mock_server_list.return_value = [
            server_info(id='123', name='name123', status='ACTIVE'),
            server_info(id='456', name='name456', status='ACTIVE')]
        self.assertEqual([Resource('OS::Nova::Server', '123', 'name123'),
                          Resource('OS::Nova::Server', '456', 'name456')],
                         plugin.list_resources(self._context))

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    @mock.patch.object(servers.ServerManager, 'get')
    def test_show_resource(self, mock_server_get, mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)

        server_info = collections.namedtuple('server_info', ['id', 'name',
                                                             'status'])
        mock_server_get.return_value = server_info(id='123', name='name123',
                                                   status='ACTIVE')
        self.assertEqual(Resource('OS::Nova::Server', '123', 'name123'),
                         plugin.show_resource(self._context, '123'))

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    @mock.patch.object(servers.ServerManager, 'list')
    def test_get_dependent_resources(self, mock_server_list,
                                     mock_generate_session):
        plugin = ServerProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)

        server_info = collections.namedtuple('server_info', ['id', 'name',
                                                             'status'])
        mock_server_list.return_value = [
            server_info(id='123', name='name123', status='ACTIVE'),
            server_info(id='456', name='name456', status='ACTIVE')]
        self.assertEqual([Resource('OS::Nova::Server', '123', 'name123'),
                          Resource('OS::Nova::Server', '456', 'name456')],
                         plugin.get_dependent_resources(self._context, None))



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_serialization import jsonutils

from karbor import objects
from karbor.tests.unit import fake_plan
from karbor.tests.unit import objects as test_objects


class TestPlan(test_objects.BaseObjectsTestCase):
    @staticmethod
    def _compare(test, db, obj):
        db = {k: v for k, v in db.items()
              if not k.endswith('resources')}
        test_objects.BaseObjectsTestCase._compare(test, db, obj)

    @mock.patch('karbor.objects.Plan.get_by_id')
    def test_get_by_id(self, plan_get):
        db_plan = fake_plan.fake_db_plan()
        plan_get.return_value = db_plan
        plan = objects.Plan.get_by_id(self.context, "1")
        plan_get.assert_called_once_with(self.context, "1")
        self._compare(self, db_plan, plan)

    @mock.patch('karbor.db.sqlalchemy.api.plan_create')
    def test_create(self, plan_create):
        db_plan = fake_plan.fake_db_plan()
        plan_create.return_value = db_plan
        plan = objects.Plan(context=self.context)
        plan.create()
        self.assertEqual(db_plan['id'], plan.id)

    @mock.patch('karbor.db.sqlalchemy.api.plan_update')
    def test_save(self, plan_update):
        db_plan = fake_plan.fake_db_plan()
        plan = objects.Plan._from_db_object(self.context,
                                            objects.Plan(), db_plan)
        plan.name = 'planname'
        plan.save()
        plan_update.assert_called_once_with(self.context, plan.id,
                                            {'name': 'planname'})

    @mock.patch('karbor.db.sqlalchemy.api.plan_resources_update',
                return_value=[
                    {'resource_id': 'key1',
                     "resource_type": "value1",
                     "extra_info": "{'availability_zone': 'az1'}"}
                ])
    @mock.patch('karbor.db.sqlalchemy.api.plan_update')
    def test_save_with_resource(self, plan_update, resource_update):
        db_plan = fake_plan.fake_db_plan()
        plan = objects.Plan._from_db_object(self.context,
                                            objects.Plan(), db_plan)
        plan.name = 'planname'
        plan.resources = [{'id': 'key1',
                           "type": "value1",
                           "extra_info": "{'availability_zone': 'az1'}"}]
        self.assertEqual({'name': 'planname',
                          'resources': [{'id': 'key1',
                                         "type": "value1",
                                         "extra_info":
                                             "{'availability_zone': 'az1'}"}]},
                         plan.obj_get_changes())
        plan.save()
        plan_update.assert_called_once_with(self.context, plan.id,
                                            {'name': 'planname'})
        resource_update.assert_called_once_with(
            self.context, plan.id, [{'id': 'key1', "type": "value1",
                                     "extra_info":
                                         "{'availability_zone': 'az1'}"}])

    @mock.patch('karbor.db.sqlalchemy.api.plan_destroy')
    def test_destroy(self, plan_destroy):
        db_plan = fake_plan.fake_db_plan()
        plan = objects.Plan._from_db_object(self.context,
                                            objects.Plan(), db_plan)
        plan.destroy()
        self.assertTrue(plan_destroy.called)
        admin_context = plan_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)

    def test_parameters(self):
        db_plan = fake_plan.fake_db_plan()
        plan = objects.Plan._from_db_object(self.context,
                                            objects.Plan(), db_plan)
        self.assertEqual(plan.parameters,
                         jsonutils.loads(fake_plan.db_plan['parameters']))

    def test_obj_fields(self):
        plan = objects.Plan(context=self.context, id="2", name="testname")
        self.assertEqual(['plan_resources'], plan.obj_extra_fields)
        self.assertEqual('testname', plan.name)
        self.assertEqual('2', plan.id)

    def test_obj_field_status(self):
        plan = objects.Plan(context=self.context,
                            status='suspending')
        self.assertEqual('suspending', plan.status)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import access
from keystoneauth1.identity import access as access_plugin
from keystoneauth1 import loading
from keystoneauth1 import session as keystone_session
from keystoneclient.v3 import client as kc_v3
from oslo_config import cfg
from oslo_log import log as logging

from karbor import exception
from karbor import utils


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

# the config of trustee is like:
# [trustee]
# auth_type = password
# auth_url = http://192.168.1.2:35357
# username = karbor
# password = password
# user_domain_id = default
KEYSTONECLIENT_VERSION = (3, 0)
TRUSTEE_CONF_GROUP = 'trustee'
loading.register_auth_conf_options(CONF, TRUSTEE_CONF_GROUP)
CONF.import_group('keystone_authtoken',
                  'keystonemiddleware.auth_token')


class KarborKeystonePlugin(object):
    """Contruct a keystone client plugin with karbor user

       to offer the following functions:

       1. get the endpoint of service, such as nova, cinder
       2. create trust to karbor
    """

    def __init__(self):
        self._client = None
        self._auth_uri = None
        self._karbor_user_id = None
        auth_plugin = self._get_karbor_auth_plugin()
        # set the project which karbor belongs to
        auth_plugin._project_name = "service"
        auth_plugin._project_domain_id = "default"
        self._service_auth_plugin = auth_plugin

    @property
    def karbor_user_id(self):
        if not self._karbor_user_id:
            lcfg = CONF[TRUSTEE_CONF_GROUP]
            self._karbor_user_id = self._get_service_user(
                lcfg.username, lcfg.user_domain_id)
        return self._karbor_user_id

    @property
    def client(self):
        if not self._client:
            self._client = self._get_keystone_client(self.service_auth_plugin)
        return self._client

    @property
    def auth_uri(self):
        if not self._auth_uri:
            try:
                self._auth_uri = utils.get_auth_uri()
            except Exception:
                msg = 'get keystone auth url failed'
                raise exception.AuthorizationFailure(obj=msg)
        return self._auth_uri

    @property
    def service_auth_plugin(self):
        return self._service_auth_plugin

    def get_service_endpoint(self, service_name, service_type,
                             region_id, interface='public'):
        try:
            service = self.client.services.list(
                name=service_name,
                service_type=service_type,
                base_url=self.auth_uri)

            endpoint = self.client.endpoints.list(
                service=service[0],
                interface=interface,
                region_id=region_id,
                base_url=self.auth_uri)

            return endpoint[0].url if endpoint else None

        except Exception:
            msg = ('get service(%s) endpoint failed' % service_name)
            raise exception.AuthorizationFailure(obj=msg)

    def create_user_auth_plugin(self, context):
        if not context.auth_token_info:
            msg = ("user=%s, project=%s" % (context.user_id,
                                            context.project_id))
            raise exception.AuthorizationFailure(obj=msg)

        auth_ref = access.create(body=context.auth_token_info,
                                 auth_token=context.auth_token)
        return access_plugin.AccessInfoPlugin(
            auth_url=self.auth_uri, auth_ref=auth_ref)

    def create_trust_to_karbor(self, context):
        l_kc_v3 = self._get_keystone_client(
            self.create_user_auth_plugin(context))
        keystone_trust_url = self.auth_uri + 'OS-TRUST'
        try:
            trust = l_kc_v3.trusts.create(trustor_user=context.user_id,
                                          trustee_user=self.karbor_user_id,
                                          project=context.project_id,
                                          impersonation=True,
                                          role_names=context.roles,
                                          base_url=keystone_trust_url)
            return trust.id

        except Exception as e:
            raise exception.AuthorizationFailure(obj=str(e))

    def delete_trust_to_karbor(self, trust_id):
        auth_plugin = self._get_karbor_auth_plugin(trust_id)
        client = self._get_keystone_client(auth_plugin)
        client.trusts.delete(trust_id)

    def create_trust_session(self, trust_id):
        auth_plugin = self._get_karbor_auth_plugin(trust_id)
        cafile = cfg.CONF.keystone_authtoken.cafile
        return keystone_session.Session(
            auth=auth_plugin, verify=False if
            CONF.keystone_authtoken.insecure else cafile)

    def _get_service_user(self, user_name, user_domain_id):
        try:
            users = self.client.users.list(
                name=user_name,
                domain=user_domain_id)

            return users[0].id if users else None

        except Exception as e:
            raise exception.AuthorizationFailure(obj=e)

    def _get_karbor_auth_plugin(self, trust_id=None):
        auth_plugin = loading.load_auth_from_conf_options(
            CONF, TRUSTEE_CONF_GROUP, trust_id=trust_id)

        if not auth_plugin:
            LOG.warning('Please add the trustee credentials you need to the'
                        ' %s section of your karbor.conf file.',
                        TRUSTEE_CONF_GROUP)
            raise exception.AuthorizationFailure(obj=TRUSTEE_CONF_GROUP)

        return auth_plugin

    def _get_keystone_client(self, auth_plugin):
        cafile = cfg.CONF.keystone_authtoken.cafile
        try:
            l_session = keystone_session.Session(
                auth=auth_plugin, verify=False if
                CONF.keystone_authtoken.insecure else cafile)
            return kc_v3.Client(version=KEYSTONECLIENT_VERSION,
                                session=l_session)
        except Exception:
            msg = ('create keystone client failed.cafile:(%s)' % cafile)
            raise exception.AuthorizationFailure(obj=msg)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
time format base class
"""

import abc
import six


@six.add_metaclass(abc.ABCMeta)
class TimeFormat(object):

    def __init__(self, start_time, pattern):
        """Initiate time format

        :param start_time: The time points after the start_time are valid
        :param pattern: The pattern of the time

        When the start_time and pattern are specified, the time points
        can be calculated and are immutable.
        """
        pass

    @classmethod
    @abc.abstractmethod
    def check_time_format(cls, pattern):
        """Check time format

        :param pattern: The pattern of the time
        """
        pass

    @abc.abstractmethod
    def compute_next_time(self, current_time):
        """Compute next time

        :param current_time: the time before the next time
        """
        pass

    @abc.abstractmethod
    def get_min_interval(self):
        """Get minimum interval of two adjacent time points"""
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utilities and helper functions."""
import ast
import os
import webob.exc

from keystoneclient import discover as ks_discover
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
from oslo_utils import strutils
from oslo_utils import timeutils

from karbor import exception
from karbor.i18n import _
from stevedore import driver

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def find_config(config_path):
    """Find a configuration file using the given hint.

    :param config_path: Full or relative path to the config.
    :returns: Full path of the config, if it exists.
    :raises: `karbor.exception.ConfigNotFound`

    """
    possible_locations = [
        config_path,
        os.path.join("/var/lib/karbor", "etc", "karbor", config_path),
        os.path.join("/var/lib/karbor", "etc", config_path),
        os.path.join("/var/lib/karbor", config_path),
        "/etc/karbor/%s" % config_path,
    ]

    for path in possible_locations:
        if os.path.exists(path):
            return os.path.abspath(path)

    raise exception.ConfigNotFound(path=os.path.abspath(config_path))


def service_is_up(service):
    """Check whether a service is up based on last heartbeat."""
    last_heartbeat = service['updated_at'] or service['created_at']

    elapsed = (timeutils.utcnow() - last_heartbeat).total_seconds()
    return abs(elapsed) <= CONF.service_down_time


def remove_invalid_filter_options(context, filters,
                                  allowed_search_options):
    """Remove search options that are not valid for non-admin API/context."""

    if context.is_admin:
        # Allow all options
        return
    # Otherwise, strip out all unknown options
    unknown_options = [opt for opt in filters
                       if opt not in allowed_search_options]
    bad_options = ", ".join(unknown_options)
    LOG.debug("Removing options '%s' from query.", bad_options)
    for opt in unknown_options:
        del filters[opt]


def check_filters(filters):
    for k, v in filters.items():
        try:
            filters[k] = ast.literal_eval(v)
        except (ValueError, SyntaxError):
            LOG.debug('Could not evaluate value %s, assuming string', v)


def is_valid_boolstr(val):
    """Check if the provided string is a valid bool string or not."""
    val = str(val).lower()
    return val in ('true', 'false', 'yes', 'no', 'y', 'n', '1', '0')


def get_bool_param(param_string, params):
    param = params.get(param_string, False)
    if not is_valid_boolstr(param):
        msg = _('Value %(param)s for %(param_string)s is not a '
                'boolean.') % {'param': param, 'param_string': param_string}
        raise exception.InvalidParameterValue(err=msg)

    return strutils.bool_from_string(param, strict=True)


def load_class(namespace, plugin_name):
    try:
        LOG.debug('Start load plugin %s. ', plugin_name)
        # Try to resolve plugin by name
        mgr = driver.DriverManager(namespace, plugin_name)
        return mgr.driver
    except RuntimeError as e1:
        # fallback to class name
        try:
            return importutils.import_class(plugin_name)
        except ImportError as e2:
            LOG.error("Error loading plugin by name, %s", e1)
            LOG.error("Error loading plugin by class, %s", e2)
            raise ImportError(_("Class not found."))


def load_plugin(namespace, plugin_name, *args, **kwargs):
    plugin_class = load_class(namespace, plugin_name)
    return plugin_class(*args, **kwargs)


def get_auth_uri(v3=True):
    # Look for the keystone auth_uri in the configuration. First we
    # check the [clients_keystone] section, and if it is not set we
    # look in [keystone_authtoken]
    if cfg.CONF.clients_keystone.auth_uri:
        discover = ks_discover.Discover(
            auth_url=cfg.CONF.clients_keystone.auth_uri)
        return discover.url_for('3.0')
    else:
        # Import auth_token to have keystone_authtoken settings setup.
        importutils.import_module('keystonemiddleware.auth_token')
        auth_uri = cfg.CONF.keystone_authtoken.auth_uri
        return auth_uri.replace('v2.0', 'v3') if auth_uri and v3 else auth_uri


def validate_integer(value, name, min_value=None, max_value=None):
    """Make sure that value is a valid integer, potentially within range.

    :param value: the value of the integer
    :param name: the name of the integer
    :param min_length: the min_length of the integer
    :param max_length: the max_length of the integer
    :returns: integer
    """
    try:
        value = int(value)
    except (TypeError, ValueError, UnicodeEncodeError):
        raise webob.exc.HTTPBadRequest(explanation=(
            _('%s must be an integer.') % name))

    if min_value is not None and value < min_value:
        raise webob.exc.HTTPBadRequest(
            explanation=(_('%(value_name)s must be >= %(min_value)d') %
                         {'value_name': name, 'min_value': min_value}))
    if max_value is not None and value > max_value:
        raise webob.exc.HTTPBadRequest(
            explanation=(_('%(value_name)s must be <= %(max_value)d') %
                         {'value_name': name, 'max_value': max_value}))

    return value


def walk_class_hierarchy(clazz, encountered=None):
    """Walk class hierarchy, yielding most derived classes first."""
    if not encountered:
        encountered = []
    for subclass in clazz.__subclasses__():
        if subclass not in encountered:
            encountered.append(subclass)
            # drill down to leaves first
            for subsubclass in walk_class_hierarchy(subclass, encountered):
                yield subsubclass
            yield subclass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from collections import namedtuple

from karbor.common import constants
from karbor import exception
from karbor.services.protection import graph
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


HOOKS = (
    HOOK_PRE_BEGIN,
    HOOK_PRE_FINISH,
    HOOK_MAIN,
    HOOK_COMPLETE
) = (
    'on_prepare_begin',
    'on_prepare_finish',
    'on_main',
    'on_complete'
)

ResourceHooks = namedtuple('ResourceHooks', [
    HOOK_PRE_BEGIN,
    HOOK_PRE_FINISH,
    HOOK_MAIN,
    HOOK_COMPLETE,
])


OPERATION_EXTRA_ARGS = {
    constants.OPERATION_RESTORE: ['heat_template', 'restore'],
}


def noop_handle(*args, **kwargs):
    pass


class ResourceFlowGraphWalkerListener(graph.GraphWalkerListener):
    def __init__(self, resource_flow, operation_type, context, parameters,
                 plugins, workflow_engine):
        super(ResourceFlowGraphWalkerListener, self).__init__()
        self.operation_type = operation_type
        self.context = context
        self.parameters = parameters or {}
        self.plugins = plugins
        self.workflow_engine = workflow_engine
        self.flow = resource_flow

        self.node_tasks = {}
        self.task_stack = []
        self.current_resource = None

    def _create_hook_tasks(self, operation_obj, resource):
        pre_begin_task = self._create_hook_task(operation_obj, resource,
                                                HOOK_PRE_BEGIN)
        pre_finish_task = self._create_hook_task(operation_obj, resource,
                                                 HOOK_PRE_FINISH)
        main_task = self._create_hook_task(operation_obj, resource,
                                           HOOK_MAIN)
        post_task = self._create_hook_task(operation_obj, resource,
                                           HOOK_COMPLETE)

        return ResourceHooks(pre_begin_task, pre_finish_task, main_task,
                             post_task)

    def _create_hook_task(self, operation_obj, resource, hook_type):
        method = getattr(operation_obj, hook_type, noop_handle)
        assert callable(method), (
            'Resource {} method "{}" is not callable'
        ).format(resource.type, hook_type)

        task_name = "{operation_type}_{hook_type}_{type}_{id}".format(
            type=resource.type,
            id=resource.id,
            hook_type=hook_type,
            operation_type=self.operation_type,
        )

        parameters = {}
        parameters.update(self.parameters.get(resource.type, {}))
        resource_id = '{}#{}'.format(resource.type, resource.id)
        parameters.update(self.parameters.get(resource_id, {}))
        injects = {
            'context': self.context,
            'parameters': parameters,
            'resource': resource,
        }
        requires = OPERATION_EXTRA_ARGS.get(self.operation_type)
        task = self.workflow_engine.create_task(method,
                                                name=task_name,
                                                inject=injects,
                                                requires=requires)
        return task

    def on_node_enter(self, node, already_visited):
        resource = node.value
        LOG.debug(
            "Enter node (type: %(type)s id: %(id)s visited: %(visited)s)",
            {"type": resource.type, "id": resource.id, "visited":
             already_visited}
        )
        self.current_resource = resource
        if already_visited:
            self.task_stack.append(self.node_tasks[resource.id])
            return

        if resource.type not in self.plugins:
            raise exception.ProtectionPluginNotFound(type=resource.type)

        protection_plugin = self.plugins[resource.type]
        operation_getter_name = 'get_{}_operation'.format(self.operation_type)
        operation_getter = getattr(protection_plugin, operation_getter_name)
        assert callable(operation_getter)
        operation_obj = operation_getter(resource)
        hooks = self._create_hook_tasks(operation_obj, resource)
        LOG.debug("added operation %s hooks", self.operation_type)
        self.node_tasks[resource.id] = hooks
        self.task_stack.append(hooks)
        self.workflow_engine.add_tasks(self.flow, hooks.on_prepare_begin,
                                       hooks.on_prepare_finish, hooks.on_main,
                                       hooks.on_complete)
        self.workflow_engine.link_task(self.flow, hooks.on_prepare_begin,
                                       hooks.on_prepare_finish)
        self.workflow_engine.link_task(self.flow, hooks.on_prepare_finish,
                                       hooks.on_main)
        self.workflow_engine.link_task(self.flow, hooks.on_main,
                                       hooks.on_complete)

    def on_node_exit(self, node):
        resource = node.value
        LOG.debug(
            "Exit node (type: %(type)s id: %(id)s)",
            {"type": resource.type, "id": resource.id}
        )
        child_hooks = self.task_stack.pop()
        if len(self.task_stack) > 0:
            parent_hooks = self.task_stack[-1]
            self.workflow_engine.link_task(self.flow,
                                           parent_hooks.on_prepare_begin,
                                           child_hooks.on_prepare_begin)
            self.workflow_engine.link_task(self.flow,
                                           child_hooks.on_prepare_finish,
                                           parent_hooks.on_prepare_finish)
            self.workflow_engine.link_task(self.flow, child_hooks.on_complete,
                                           parent_hooks.on_complete)


def build_resource_flow(operation_type, context, workflow_engine,
                        plugins, resource_graph, parameters):
    LOG.info("Build resource flow for operation %s", operation_type)

    resource_graph_flow = workflow_engine.build_flow(
        'ResourceGraphFlow_{}'.format(operation_type),
        'graph',
    )
    resource_walker = ResourceFlowGraphWalkerListener(resource_graph_flow,
                                                      operation_type,
                                                      context,
                                                      parameters,
                                                      plugins,
                                                      workflow_engine)
    walker = graph.GraphWalker()
    walker.register_listener(resource_walker)
    LOG.debug("Starting resource graph walk (operation %s)", operation_type)
    walker.walk_graph(resource_graph)
    LOG.debug("Finished resource graph walk (operation %s)", operation_type)
    return resource_graph_flow



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

OPTIONS_SCHEMA = {
    "title": "Server Protection Options",
    "type": "object",
    "properties": {},
    "required": []
}

RESTORE_SCHEMA = {
    "title": "Server Protection Restore",
    "type": "object",
    "properties": {
        "restore_name": {
            "type": "string",
            "title": "Restore Server Name",
            "description": "The name of the restore server",
        },
    },
    "required": ["restore_name"]
}

# TODO(luobin)
SAVED_INFO_SCHEMA = {
    "title": "Server Protection Saved Info",
    "type": "object",
    "properties": {
        "attach_metadata": {
            "type": "object",
            "title": "Attached Volume Metadata",
            "description": "The devices of attached volumes"
        },
        "snapshot_metadata": {
            "type": "object",
            "title": "Snapshot Metadata",
            "description": "The metadata of snapshot"
        },
    },
    "required": ["attached_metadata", "snapshot_metadata"]
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from novaclient import client as nc
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import config
from karbor import exception
from karbor.services.protection.clients import utils

LOG = logging.getLogger(__name__)

SERVICE = "nova"
nova_client_opts = [
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the nova endpoint. '
                    '<endpoint_url>'),
    cfg.StrOpt(SERVICE + '_catalog_info',
               default='compute:nova:publicURL',
               help='Info to match when looking for nova in the service '
               'catalog. Format is: separated values of the form: '
               '<service_type>:<service_name>:<endpoint_type> - '
               'Only used if nova_endpoint is unset'),
    cfg.StrOpt(SERVICE + '_ca_cert_file',
               help='Location of the CA certificate file '
                    'to use for client requests in SSL connections.'),
    cfg.BoolOpt(SERVICE + '_auth_insecure',
                default=False,
                help='Bypass verification of server certificate when '
                     'making SSL connection to Nova.'),
]

CONFIG_GROUP = '%s_client' % SERVICE
CONF = cfg.CONF
CONF.register_opts(config.service_client_opts, group=CONFIG_GROUP)
CONF.register_opts(nova_client_opts, group=CONFIG_GROUP)
CONF.set_default('service_name', 'nova', CONFIG_GROUP)
CONF.set_default('service_type', 'compute', CONFIG_GROUP)

NOVACLIENT_VERSION = '2'


def create(context, conf, **kwargs):
    conf.register_opts(nova_client_opts, group=CONFIG_GROUP)

    client_config = conf[CONFIG_GROUP]
    url = utils.get_url(SERVICE, context, client_config,
                        append_project_fmt='%(url)s/%(project)s', **kwargs)
    LOG.debug('Creating nova client with url %s.', url)

    extensions = nc.discover_extensions(NOVACLIENT_VERSION)
    session = kwargs.get('session')
    if session is None:
        LOG.error('Creating nova client failed with url %s.', url)
        raise exception.InvalidParameterValue(
            err="The parameter session is None.")

    return nc.Client(NOVACLIENT_VERSION, extensions=extensions,
                     session=kwargs.get('session'), endpoint_override=url)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import config
from karbor.common import karbor_keystone_plugin
from karborclient import client as karbor_client

LOG = logging.getLogger(__name__)

CONFIG_GROUP = 'karbor_client'
CONF = cfg.CONF
CONF.register_opts(config.service_client_opts,
                   group=CONFIG_GROUP)


def get_karbor_endpoint():
    try:
        sc_cfg = CONF[CONFIG_GROUP]
        kc_plugin = karbor_keystone_plugin.KarborKeystonePlugin()
        url = kc_plugin.get_service_endpoint(
            sc_cfg.service_name, sc_cfg.service_type,
            sc_cfg.region_id, sc_cfg.interface)

        return url.replace("$(", "%(")
    except Exception:
        raise


def create(context, **kwargs):
    endpoint = kwargs.get('endpoint')
    if not endpoint:
        endpoint = get_karbor_endpoint() % {"project_id": context.project_id}

    LOG.debug("Creating karbor client with url %s.", endpoint)

    sc_cfg = CONF[CONFIG_GROUP]
    args = {
        'version': sc_cfg.version,
        'endpoint': endpoint,
        'token': context.auth_token,
        'cacert': sc_cfg.ca_cert_file,
        'insecure': sc_cfg.auth_insecure,
    }

    return karbor_client.Client(**args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Protection Service
"""

from datetime import datetime
from eventlet import greenpool
from eventlet import greenthread
import six

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging

from oslo_utils import uuidutils

from karbor.common import constants
from karbor import exception
from karbor.i18n import _
from karbor import manager
from karbor.resource import Resource
from karbor.services.protection.flows import worker as flow_manager
from karbor.services.protection.protectable_registry import ProtectableRegistry
from karbor import utils

LOG = logging.getLogger(__name__)

protection_manager_opts = [
    cfg.StrOpt('provider_registry',
               default='provider-registry',
               help='the provider registry'),
    cfg.IntOpt('max_concurrent_operations',
               default=0,
               help='number of maximum concurrent operation (protect, restore,'
                    ' delete) flows. 0 means no hard limit'
               )
]

CONF = cfg.CONF
CONF.register_opts(protection_manager_opts)

PROVIDER_NAMESPACE = 'karbor.provider'


class ProtectionManager(manager.Manager):
    """karbor Protection Manager."""

    RPC_API_VERSION = '1.0'

    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, service_name=None,
                 *args, **kwargs):
        super(ProtectionManager, self).__init__(*args, **kwargs)
        provider_reg = CONF.provider_registry
        self.provider_registry = utils.load_plugin(PROVIDER_NAMESPACE,
                                                   provider_reg)
        self.protectable_registry = ProtectableRegistry()
        self.protectable_registry.load_plugins()
        self.worker = flow_manager.Worker()
        self._greenpool = None
        self._greenpool_size = CONF.max_concurrent_operations
        if self._greenpool_size != 0:
            self._greenpool = greenpool.GreenPool(self._greenpool_size)

    def _spawn(self, func, *args, **kwargs):
        if self._greenpool is not None:
            return self._greenpool.spawn_n(func, *args, **kwargs)
        else:
            return greenthread.spawn_n(func, *args, **kwargs)

    def init_host(self, **kwargs):
        """Handle initialization if this is a standalone service"""
        # TODO(wangliuan)
        LOG.info("Starting protection service")

    @messaging.expected_exceptions(exception.InvalidPlan,
                                   exception.ProviderNotFound,
                                   exception.FlowError)
    def protect(self, context, plan, checkpoint_properties=None):
        """create protection for the given plan

        :param plan: Define that protection plan should be done
        """

        LOG.info("Starting protection service:protect action")
        LOG.debug("protecting: %s checkpoint_properties:%s",
                  plan, checkpoint_properties)

        if not plan:
            raise exception.InvalidPlan(
                reason=_('the protection plan is None'))
        provider_id = plan.get('provider_id', None)
        plan_id = plan.get('id', None)
        provider = self.provider_registry.show_provider(provider_id)
        checkpoint_collection = provider.get_checkpoint_collection()
        try:
            checkpoint = checkpoint_collection.create(plan,
                                                      checkpoint_properties)
        except Exception as e:
            LOG.exception("Failed to create checkpoint, plan: %s", plan_id)
            exc = exception.FlowError(flow="protect",
                                      error="Error creating checkpoint")
            six.raise_from(exc, e)
        try:
            flow = self.worker.get_flow(
                context=context,
                protectable_registry=self.protectable_registry,
                operation_type=constants.OPERATION_PROTECT,
                plan=plan,
                provider=provider,
                checkpoint=checkpoint)
        except Exception as e:
            LOG.exception("Failed to create protection flow, plan: %s",
                          plan_id)
            raise exception.FlowError(
                flow="protect",
                error=e.msg if hasattr(e, 'msg') else 'Internal error')
        self._spawn(self.worker.run_flow, flow)
        return checkpoint.id

    @messaging.expected_exceptions(exception.ProviderNotFound,
                                   exception.CheckpointNotFound,
                                   exception.CheckpointNotAvailable,
                                   exception.FlowError,
                                   exception.InvalidInput)
    def restore(self, context, restore, restore_auth):
        LOG.info("Starting restore service:restore action")

        checkpoint_id = restore["checkpoint_id"]
        provider_id = restore["provider_id"]
        provider = self.provider_registry.show_provider(provider_id)
        if not provider:
            raise exception.ProviderNotFound(provider_id=provider_id)

        self.validate_restore_parameters(restore, provider)

        checkpoint_collection = provider.get_checkpoint_collection()
        checkpoint = checkpoint_collection.get(checkpoint_id)

        if checkpoint.status != constants.CHECKPOINT_STATUS_AVAILABLE:
            raise exception.CheckpointNotAvailable(
                checkpoint_id=checkpoint_id)

        try:
            flow = self.worker.get_flow(
                context=context,
                operation_type=constants.OPERATION_RESTORE,
                checkpoint=checkpoint,
                provider=provider,
                restore=restore,
                restore_auth=restore_auth)
        except Exception:
            LOG.exception("Failed to create restore flow checkpoint: %s",
                          checkpoint_id)
            raise exception.FlowError(
                flow="restore",
                error=_("Failed to create flow"))
        self._spawn(self.worker.run_flow, flow)

    def validate_restore_parameters(self, restore, provider):
        parameters = restore["parameters"]
        if not parameters:
            return
        restore_schema = provider.extended_info_schema.get(
            "restore_schema", None)
        if restore_schema is None:
            msg = _("The restore schema of plugin must be provided.")
            raise exception.InvalidInput(reason=msg)
        for resource_key, parameter_value in parameters.items():
            if "#" in resource_key:
                resource_type, resource_id = resource_key.split("#")
                if not uuidutils.is_uuid_like(resource_id):
                    msg = _("The resource_id must be a uuid.")
                    raise exception.InvalidInput(reason=msg)
            else:
                resource_type = resource_key
            if (resource_type not in constants.RESOURCE_TYPES) or (
                    resource_type not in restore_schema):
                msg = _("The key of restore parameters is invalid.")
                raise exception.InvalidInput(reason=msg)
            properties = restore_schema[resource_type]["properties"]
            if not set(parameter_value.keys()).issubset(
                    set(properties.keys())):
                msg = _("The restore property of restore parameters "
                        "is invalid.")
                raise exception.InvalidInput(reason=msg)

    def delete(self, context, provider_id, checkpoint_id):
        LOG.info("Starting protection service:delete action")
        LOG.debug('provider_id :%s checkpoint_id:%s', provider_id,
                  checkpoint_id)
        provider = self.provider_registry.show_provider(provider_id)
        try:
            checkpoint_collection = provider.get_checkpoint_collection()
            checkpoint = checkpoint_collection.get(checkpoint_id)
        except Exception:
            LOG.error("get checkpoint failed, checkpoint_id:%s",
                      checkpoint_id)
            raise exception.InvalidInput(
                reason=_("Invalid checkpoint_id or provider_id"))

        if checkpoint.status not in [
            constants.CHECKPOINT_STATUS_AVAILABLE,
            constants.CHECKPOINT_STATUS_ERROR,
        ]:
            raise exception.CheckpointNotBeDeleted(
                checkpoint_id=checkpoint_id)
        checkpoint.status = constants.CHECKPOINT_STATUS_DELETING
        checkpoint.commit()

        try:
            flow = self.worker.get_flow(
                context=context,
                operation_type=constants.OPERATION_DELETE,
                checkpoint=checkpoint,
                provider=provider)
        except Exception:
            LOG.exception("Failed to create delete checkpoint flow,"
                          "checkpoint:%s.", checkpoint_id)
            raise exception.KarborException(_(
                "Failed to create delete checkpoint flow."
            ))
        self._spawn(self.worker.run_flow, flow)

    def start(self, plan):
        # TODO(wangliuan)
        pass

    def suspend(self, plan):
        # TODO(wangliuan)
        pass

    @messaging.expected_exceptions(exception.ProviderNotFound,
                                   exception.CheckpointNotFound)
    def list_checkpoints(self, context, provider_id, marker=None, limit=None,
                         sort_keys=None, sort_dirs=None, filters=None):
        LOG.info("Starting list checkpoints. provider_id:%s", provider_id)
        plan_id = filters.get("plan_id", None)
        start_date = None
        end_date = None
        if filters.get("start_date", None):
            start_date = datetime.strptime(
                filters.get("start_date"), "%Y-%m-%d")
        if filters.get("end_date", None):
            end_date = datetime.strptime(
                filters.get("end_date"), "%Y-%m-%d")
        sort_dir = None if sort_dirs is None else sort_dirs[0]
        provider = self.provider_registry.show_provider(provider_id)
        checkpoint_ids = provider.list_checkpoints(
            provider_id, limit=limit, marker=marker, plan_id=plan_id,
            start_date=start_date, end_date=end_date, sort_dir=sort_dir)
        checkpoints = []
        for checkpoint_id in checkpoint_ids:
            checkpoint = provider.get_checkpoint(checkpoint_id)
            checkpoints.append(checkpoint.to_dict())
        return checkpoints

    @messaging.expected_exceptions(exception.ProviderNotFound,
                                   exception.CheckpointNotFound)
    def show_checkpoint(self, context, provider_id, checkpoint_id):
        provider = self.provider_registry.show_provider(provider_id)

        checkpoint = provider.get_checkpoint(checkpoint_id)
        return checkpoint.to_dict()

    def list_protectable_types(self, context):
        LOG.info("Start to list protectable types.")
        return self.protectable_registry.list_resource_types()

    @messaging.expected_exceptions(exception.ProtectableTypeNotFound)
    def show_protectable_type(self, context, protectable_type):
        LOG.info("Start to show protectable type %s", protectable_type)

        plugin = self.protectable_registry.get_protectable_resource_plugin(
            protectable_type)
        if not plugin:
            raise exception.ProtectableTypeNotFound(
                protectable_type=protectable_type)

        dependents = []
        for t in self.protectable_registry.list_resource_types():
            if t == protectable_type:
                continue

            p = self.protectable_registry.get_protectable_resource_plugin(t)
            if p and protectable_type in p.get_parent_resource_types():
                dependents.append(t)

        return {
            'name': plugin.get_resource_type(),
            "dependent_types": dependents
        }

    @messaging.expected_exceptions(exception.ListProtectableResourceFailed)
    def list_protectable_instances(self, context,
                                   protectable_type=None,
                                   marker=None,
                                   limit=None,
                                   sort_keys=None,
                                   sort_dirs=None,
                                   filters=None,
                                   parameters=None):

        LOG.info("Start to list protectable instances of type: %s",
                 protectable_type)

        try:
            resource_instances = self.protectable_registry.list_resources(
                context, protectable_type, parameters)
        except exception.ListProtectableResourceFailed as err:
            LOG.error("List resources of type %(type)s failed: %(err)s",
                      {'type': protectable_type, 'err': six.text_type(err)})
            raise

        result = []
        for resource in resource_instances:
            result.append(dict(id=resource.id, name=resource.name,
                               extra_info=resource.extra_info))

        return result

    @messaging.expected_exceptions(exception.ListProtectableResourceFailed)
    def show_protectable_instance(self, context, protectable_type,
                                  protectable_id, parameters=None):
        LOG.info("Start to show protectable instance of type: %s",
                 protectable_type)

        registry = self.protectable_registry
        try:
            resource_instance = registry.show_resource(
                context,
                protectable_type,
                protectable_id,
                parameters=parameters
            )
        except exception.ListProtectableResourceFailed as err:
            LOG.error("Show resources of type %(type)s id %(id)s "
                      "failed: %(err)s",
                      {'type': protectable_type,
                       'id': protectable_id,
                       'err': six.text_type(err)})
            raise

        return resource_instance.to_dict()

    @messaging.expected_exceptions(exception.ListProtectableResourceFailed)
    def list_protectable_dependents(self, context,
                                    protectable_id,
                                    protectable_type):
        LOG.info("Start to list dependents of resource (type:%(type)s, "
                 "id:%(id)s)",
                 {'type': protectable_type,
                  'id': protectable_id})

        parent_resource = Resource(type=protectable_type, id=protectable_id,
                                   name="")

        registry = self.protectable_registry
        try:
            dependent_resources = registry.fetch_dependent_resources(
                context, parent_resource)
        except exception.ListProtectableResourceFailed as err:
            LOG.error("List dependent resources of (%(res)s) failed: %(err)s",
                      {'res': parent_resource,
                       'err': six.text_type(err)})
            raise

        return [resource.to_dict() for resource in dependent_resources]

    def list_providers(self, context, marker=None, limit=None,
                       sort_keys=None, sort_dirs=None, filters=None):
        return self.provider_registry.list_providers(marker=marker,
                                                     limit=limit,
                                                     sort_keys=sort_keys,
                                                     sort_dirs=sort_dirs,
                                                     filters=filters)

    @messaging.expected_exceptions(exception.ProviderNotFound)
    def show_provider(self, context, provider_id):
        provider = self.provider_registry.show_provider(provider_id)
        response = {'id': provider.id,
                    'name': provider.name,
                    'description': provider.description,
                    'extended_info_schema': provider.extended_info_schema,
                    }
        return response



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import abstractmethod
from datetime import datetime
from datetime import timedelta
from oslo_log import log as logging
from oslo_utils import timeutils

from karbor.common import constants
from karbor import context
from karbor import objects
from karbor.services.operationengine.engine.executors import base


LOG = logging.getLogger(__name__)


class ScheduledOperationExecutor(base.BaseExecutor):

    _CHECK_ITEMS = {
        'is_waiting': 'is_waiting',
        'is_canceled': 'is_canceled'
    }

    def execute_operation(self, operation_id, triggered_time,
                          expect_start_time, window_time, **kwargs):

        if self._check_operation(operation_id, self._CHECK_ITEMS.values()):
            LOG.warning("Execute operation(%s), it can't be executed",
                        operation_id)
            return

        end_time_for_run = expect_start_time + timedelta(seconds=window_time)
        ret = self._update_operation_state(
            operation_id,
            {'state': constants.OPERATION_STATE_RUNNING,
             'end_time_for_run': end_time_for_run})
        if not ret:
            return

        param = {
            'operation_id': operation_id,
            'triggered_time': triggered_time,
            'expect_start_time': expect_start_time,
            'window_time': window_time,
            'run_type': constants.OPERATION_RUN_TYPE_EXECUTE
        }
        self._execute_operation(operation_id, self._run_operation, param)

    def resume_operation(self, operation_id, **kwargs):
        end_time = kwargs.get('end_time_for_run')
        now = datetime.utcnow()
        if not isinstance(end_time, datetime) or now > end_time:
            return

        window = int(timeutils.delta_seconds(now, end_time))
        param = {
            'operation_id': operation_id,
            'triggered_time': now,
            'expect_start_time': now,
            'window_time': window,
            'run_type': constants.OPERATION_RUN_TYPE_RESUME
        }
        self._execute_operation(operation_id, self._run_operation, param)

    def _run_operation(self, operation_id, param):

        try:
            check_item = [self._CHECK_ITEMS['is_canceled']]
            if self._check_operation(operation_id, check_item):
                return

            try:
                operation = objects.ScheduledOperation.get_by_id(
                    context.get_admin_context(), operation_id)
            except Exception:
                LOG.exception("Run operation(%s), get operation failed",
                              operation_id)
                return

            try:
                param['user_id'] = operation.user_id
                param['project_id'] = operation.project_id

                self._operation_manager.run_operation(
                    operation.operation_type,
                    operation.operation_definition,
                    param=param)
            except Exception:
                LOG.exception("Run operation(%s) failed", operation_id)

        finally:
            self._update_operation_state(
                operation_id,
                {'state': constants.OPERATION_STATE_REGISTERED})

    def _update_operation_state(self, operation_id, updates):

        ctxt = context.get_admin_context()
        try:
            state_ref = objects.ScheduledOperationState.get_by_operation_id(
                ctxt, operation_id)
            for item, value in updates.items():
                setattr(state_ref, item, value)
            state_ref.save()
        except Exception:
            LOG.exception("Execute operation(%s), update state failed",
                          operation_id)
            return False
        return True

    @abstractmethod
    def _execute_operation(self, operation_id, funtion, param):
        pass

    @abstractmethod
    def _check_operation(self, operation_id, check_items):
        """Check whether the item in check_items happens"""
        pass




#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg

from karbor.context import RequestContext
from karbor.services.protection.clients import neutron
from karbor.tests import base


class NeutronClientTest(base.TestCase):
    def setUp(self):
        super(NeutronClientTest, self).setUp()

        self._public_url = 'http://127.0.0.1:9696'

        service_catalog = [
            {
                'endpoints': [{'publicURL': self._public_url}],
                'type': 'network',
                'name': 'neutron',
            },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    def test_create_client_by_endpoint(self):
        cfg.CONF.set_default('neutron_endpoint',
                             'http://127.0.0.1:9696',
                             'neutron_client')
        nc = neutron.create(self._context, cfg.CONF)
        self.assertEqual('http://127.0.0.1:9696', nc.httpclient.endpoint_url)

    def test_create_client_by_catalog(self):
        nc = neutron.create(self._context, cfg.CONF)
        self.assertEqual('http://127.0.0.1:9696', nc.httpclient.endpoint_url)

    @mock.patch('karbor.services.protection.clients.utils.get_url')
    @mock.patch('neutronclient.client.construct_http_client')
    def test_create_client(self, create, get_url):
        get_url.return_value = self._public_url

        client_config = cfg.CONF[neutron.CONFIG_GROUP]
        session = object()
        args = {
            'endpoint_url': self._public_url,
            'token': self._context.auth_token,
            'cacert': client_config.neutron_ca_cert_file,
            'insecure': client_config.neutron_auth_insecure,
        }

        neutron.create(self._context, cfg.CONF)
        create.assert_called_with(**args)

        neutron.create(self._context, cfg.CONF, session=session)
        create.assert_called_with(endpoint_override=self._public_url,
                                  session=session)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import six

from karbor.common import constants
from karbor import exception
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protection_plugin
from karbor.services.protection.protection_plugins.network \
    import network_plugin_schemas
from karbor.services.protection.restore_heat import HeatResource
from oslo_log import log as logging
from uuid import uuid4

LOG = logging.getLogger(__name__)


def get_network_id(cntxt):
    network_id = cntxt.project_id
    return network_id


class ProtectOperation(protection_plugin.Operation):
    _SUPPORT_RESOURCE_TYPES = [constants.NETWORK_RESOURCE_TYPE]

    def _get_resources_by_network(self, cntxt, neutron_client):
        try:
            networks = neutron_client.list_networks().get('networks')
            networks_metadata = {}

            allowed_keys = [
                'id',
                'admin_state_up',
                'availability_zone_hints',
                'description',
                'ipv4_address_scope',
                'ipv6_address_scope',
                'mtu',
                'name',
                'port_security_enabled',
                'router:external',
                'shared',
                'status',
                'subnets',
                'tags',
                'tenant_id'
                ]

            for network in networks:
                network_metadata = {
                    k: network[k] for k in network if k in allowed_keys}
                networks_metadata[network["id"]] = network_metadata
            return networks_metadata
        except Exception as e:
            LOG.exception("List all summary networks from neutron failed.")
            raise exception.GetProtectionNetworkSubResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPES,
                reason=six.text_type(e))
        else:
            return []

    def _get_resources_by_subnet(self, cntxt, neutron_client):
        try:
            subnets = neutron_client.list_subnets().get('subnets')
            subnets_metadata = {}

            allowed_keys = [
                'cidr',
                'allocation_pools',
                'description',
                'dns_nameservers',
                'enable_dhcp',
                'gateway_ip',
                'host_routes',
                'id',
                'ip_version',
                'ipv6_address_mode',
                'ipv6_ra_mode',
                'name',
                'network_id',
                'subnetpool_id',
                'tenant_id'
                ]

            for subnet in subnets:
                subnet_metadata = {
                    k: subnet[k] for k in subnet if k in allowed_keys}
                subnets_metadata[subnet["id"]] = subnet_metadata

            return subnets_metadata
        except Exception as e:
            LOG.exception("List all summary subnets from neutron failed.")
            raise exception.GetProtectionNetworkSubResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPES,
                reason=six.text_type(e))
        else:
            return []

    def _get_resources_by_port(self, cntxt, neutron_client):
        try:
            ports = neutron_client.list_ports().get('ports')
            ports_metadata = {}

            allowed_keys = [
                'admin_state_up',
                'allowed_address_pairs',
                'description',
                'device_id',
                'device_owner',
                'extra_dhcp_opts',
                'fixed_ips',
                'id',
                'mac_address',
                'name',
                'network_id',
                'port_security_enabled',
                'security_groups',
                'status',
                'tenant_id'
                ]

            for port in ports:
                port_metadata = {
                    k: port[k] for k in port if k in allowed_keys}
                ports_metadata[port["id"]] = port_metadata
            return ports_metadata
        except Exception as e:
            LOG.exception("List all summary ports from neutron failed.")
            raise exception.GetProtectionNetworkSubResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPES,
                reason=six.text_type(e))
        else:
            return []

    def _get_resources_by_router(self, cntxt, neutron_client):
        try:
            routers = neutron_client.list_routers().get('routers')
            routers_metadata = {}

            allowed_keys = [
                'admin_state_u',
                'availability_zone_hints',
                'description',
                'external_gateway_info',
                'id',
                'name',
                'routes',
                'status'
                ]

            for router in routers:
                router_metadata = {
                    k: router[k] for k in router if k in allowed_keys}
                routers_metadata[router["id"]] = router_metadata

            return routers_metadata
        except Exception as e:
            LOG.exception("List all summary routers from neutron failed.")
            raise exception.GetProtectionNetworkSubResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPES,
                reason=six.text_type(e))
        else:
            return []

    def _get_resources_by_security_group(self, cntxt, neutron_client):
        try:
            sgs = neutron_client.list_security_groups().get('security_groups')
            sgs_metadata = {}

            allowed_keys = [
                'id',
                'description',
                'name',
                'security_group_rules',
                'tenant_id'
                ]

            for sg in sgs:
                sg_metadata = {k: sg[k] for k in sg if k in allowed_keys}
                sgs_metadata[sg["id"]] = sg_metadata
            return sgs_metadata
        except Exception as e:
            LOG.exception("List all summary security_groups from neutron "
                          "failed.")
            raise exception.GetProtectionNetworkSubResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPES,
                reason=six.text_type(e))
        else:
            return []

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        network_id = get_network_id(context)
        backup_name = kwargs.get("backup_name", "karbor network backup")
        bank_section = checkpoint.get_resource_bank_section(network_id)
        neutron_client = ClientFactory.create_client("neutron", context)

        resource_definition = {"resource_id": network_id}
        resource_definition["backup_name"] = backup_name
        resource_definition["network_metadata"] = (
            self._get_resources_by_network(context, neutron_client))
        resource_definition["subnet_metadata"] = (
            self._get_resources_by_subnet(context, neutron_client))
        resource_definition["port_metadata"] = (
            self._get_resources_by_port(context, neutron_client))
        resource_definition["router_metadata"] = (
            self._get_resources_by_router(context, neutron_client))
        resource_definition["security-group_metadata"] = (
            self._get_resources_by_security_group(context, neutron_client))

        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_PROTECTING)

            # write resource_definition in bank
            bank_section.update_object("metadata", resource_definition)

            # update resource_definition backup_status
            bank_section.update_object("status",
                                       constants.CHECKPOINT_STATUS_AVAILABLE)
            LOG.info("finish backup network, network_id: %s.", network_id)
        except Exception as err:
            # update resource_definition backup_status
            LOG.error("create backup failed, network_id: %s.", network_id)
            bank_section.update_object("status",
                                       constants.CHECKPOINT_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason=err,
                resource_id=network_id,
                resource_type=self._SUPPORT_RESOURCE_TYPES)


class RestoreOperation(protection_plugin.Operation):
    def _heat_restore_networks(self, heat_template, nets_meta):
        for net_meta in nets_meta:
            net_data = nets_meta[net_meta]
            if net_data["router:external"]:
                continue
            heat_resource_id = net_data["name"]
            net_heat_resource = HeatResource(heat_resource_id,
                                             constants.NET_RESOURCE_TYPE)
            properties = {}
            properties["admin_state_up"] = net_data["admin_state_up"]
            properties["port_security_enabled"] = (
                net_data["port_security_enabled"])
            properties["shared"] = net_data["shared"]
            properties["name"] = net_data["name"]

            for key, value in properties.items():
                net_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, net_heat_resource)

    def _get_dependent_net(self, network_id, nets_meta):
        for net_meta in nets_meta:
            net_data = nets_meta[net_meta]
            if network_id == net_data["id"]:
                return net_data["name"]

    def _is_external_subnet(self, network_id, nets_meta):
        for net_meta in nets_meta:
            net_data = nets_meta[net_meta]
            if network_id == net_data["id"]:
                return net_data["router:external"]

    def _heat_restore_subnets(self, heat_template, nets_meta, subs_meta):
        for sub_meta in subs_meta:
            sub_data = subs_meta[sub_meta]

            is_ext_subnet = self._is_external_subnet(sub_data["network_id"],
                                                     nets_meta)

            if is_ext_subnet:
                continue

            heat_resource_id = sub_data["name"]
            sub_heat_resource = HeatResource(heat_resource_id,
                                             constants.SUBNET_RESOURCE_TYPE)
            properties = {}
            properties["cidr"] = sub_data["cidr"]
            properties["allocation_pools"] = sub_data["allocation_pools"]
            properties["dns_nameservers"] = sub_data["dns_nameservers"]
            properties["enable_dhcp"] = sub_data["enable_dhcp"]
            properties["gateway_ip"] = sub_data["gateway_ip"]
            properties["host_routes"] = sub_data["host_routes"]
            properties["name"] = sub_data["name"]
            properties["ip_version"] = sub_data["ip_version"]
            net_name = self._get_dependent_net(sub_data["network_id"],
                                               nets_meta)
            properties["network_id"] = (
                heat_template.get_resource_reference(net_name))
            properties["tenant_id"] = sub_data["tenant_id"]

            for key, value in properties.items():
                sub_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, sub_heat_resource)

    def _get_subnet_by_subnetid(self, subnet_id, subs_meta):
        for sub_meta in subs_meta:
            sub_data = subs_meta[sub_meta]
            if subnet_id == sub_data["id"]:
                return sub_data["name"]

        return ""

    def _get_new_fixed_ips(self, heat_template, subs_meta, fixed_ips_meta):
        new_fixed_ips = []
        for fixed_ip in fixed_ips_meta:
            properties = {}
            properties["ip_address"] = fixed_ip["ip_address"]
            subnet_name = self._get_subnet_by_subnetid(fixed_ip["subnet_id"],
                                                       subs_meta)
            properties["subnet_id"] = (
                heat_template.get_resource_reference(subnet_name))
            new_fixed_ips.append(properties)

        return new_fixed_ips

    def _heat_restore_ports(self, heat_template,
                            nets_meta, subs_meta, ports_meta):
        for port_meta in ports_meta:
            port_data = ports_meta[port_meta]
            heat_resource_id = port_data["name"]
            port_heat_resource = HeatResource(heat_resource_id,
                                              constants.PORT_RESOURCE_TYPE)

            if (port_data["device_owner"] == "network:router_interface") or (
                    port_data["device_owner"] == "network:router_gateway") or (
                    port_data["device_owner"] == "network:dhcp") or (
                    port_data["device_owner"] == "network:floatingip"):
                continue

            properties = {}
            properties["admin_state_up"] = port_data["admin_state_up"]
            properties["allowed_address_pairs"] = (
                port_data["allowed_address_pairs"])
            properties["device_id"] = port_data["device_id"]
            properties["device_owner"] = port_data["device_owner"]
            new_fixed_ips = self._get_new_fixed_ips(heat_template,
                                                    subs_meta,
                                                    port_data["fixed_ips"])
            properties["fixed_ips"] = new_fixed_ips
            properties["mac_address"] = port_data["mac_address"]
            properties["name"] = port_data["name"]
            net_name = self._get_dependent_net(port_data["network_id"],
                                               nets_meta)
            properties["network_id"] = (
                heat_template.get_resource_reference(net_name))
            properties["port_security_enabled"] = (
                port_data["port_security_enabled"])
            properties["security_groups"] = port_data["security_groups"]

            for key, value in properties.items():
                port_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, port_heat_resource)

    def _get_new_external_gateway(self, public_network_id,
                                  gateway_info, subs_meta, neutron_client):
        new_ext_gw = {}

        # get public network id
        if not public_network_id:
            networks = neutron_client.list_networks().get('networks')
            for network in networks:
                if network['router:external'] is True:
                    public_network_id = network['id']
                    break

        new_ext_gw["network"] = public_network_id
        new_ext_gw["enable_snat"] = gateway_info["enable_snat"]
        return new_ext_gw

    def _heat_restore_routers(self, public_network_id, heat_template,
                              subs_meta, routers_meta, neutron_client):
        for router_meta in routers_meta:
            router_data = routers_meta[router_meta]
            heat_resource_id = router_data["name"]
            router_heat_resource = HeatResource(heat_resource_id,
                                                constants.ROUTER_RESOURCE_TYPE)
            properties = {}
            org_external_gateway = router_data["external_gateway_info"]
            new_external_gateway = (
                self._get_new_external_gateway(public_network_id,
                                               org_external_gateway,
                                               subs_meta,
                                               neutron_client))
            properties["external_gateway_info"] = new_external_gateway
            properties["name"] = router_data["name"]

            for key, value in properties.items():
                router_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, router_heat_resource)

    def _get_router_name(self, device_id, routers_meta):
        for router_meta in routers_meta:
            router_data = routers_meta[router_meta]
            if device_id == router_data["id"]:
                return router_data["name"]

        return ""

    def _get_subnet_name_by_fixed_ips(self, fixed_ips, subs_meta):
        subnet_id = fixed_ips[0]["subnet_id"]
        for sub_meta in subs_meta:
            sub_data = subs_meta[sub_meta]
            if subnet_id == sub_data["id"]:
                return sub_data["name"]

        return ""

    def _heat_restore_routerinterfaces(self, heat_template,
                                       subs_meta, routers_meta, ports_meta):
        for port_meta in ports_meta:
            port_data = ports_meta[port_meta]
            heat_resource_id = str(uuid4())
            port_heat_resource = (
                HeatResource(heat_resource_id,
                             constants.ROUTERINTERFACE_RESOURCE_TYPE))

            if port_data["device_owner"] != "network:router_interface":
                continue

            properties = {}
            router_name = self._get_router_name(port_data["device_id"],
                                                routers_meta)
            properties["router"] = (
                heat_template.get_resource_reference(router_name))
            subnet_name = (
                self._get_subnet_name_by_fixed_ips(port_data["fixed_ips"],
                                                   subs_meta))
            properties["subnet"] = (
                heat_template.get_resource_reference(subnet_name))

            for key, value in properties.items():
                port_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, port_heat_resource)

    def _get_security_group_rules(self, security_group_rules):
        new_security_group_rules = []
        for sg in security_group_rules:
            if sg["remote_ip_prefix"] is None:
                continue

            security_group_rule = {}
            security_group_rule["direction"] = sg["direction"]
            security_group_rule["ethertype"] = sg["ethertype"]
            security_group_rule["port_range_max"] = sg["port_range_max"]
            security_group_rule["port_range_min"] = sg["port_range_min"]
            security_group_rule["protocol"] = sg["protocol"]
            security_group_rule["remote_group_id"] = sg["remote_group_id"]
            security_group_rule["remote_ip_prefix"] = sg["remote_ip_prefix"]

            if "remote_mode" in sg:
                security_group_rule["remote_mode"] = sg["remote_mode"]
            new_security_group_rules.append(security_group_rule)

        return new_security_group_rules

    def _heat_restore_securitygroups(self, heat_template, sgs_meta):
        for sg_meta in sgs_meta:
            sg_data = sgs_meta[sg_meta]

            # Skip the default securitygroups
            if sg_data["name"] == "default":
                continue

            heat_resource_id = sg_data["name"]
            sg_heat_resource = (
                HeatResource(heat_resource_id,
                             constants.SECURITYGROUP_RESOURCE_TYPE))
            properties = {}
            sg_rules = sg_data["security_group_rules"]
            properties["description"] = sg_data["description"]
            properties["name"] = sg_data["name"]
            properties["rules"] = self._get_security_group_rules(sg_rules)

            for key, value in properties.items():
                sg_heat_resource.set_property(key, value)
            heat_template.put_resource(heat_resource_id, sg_heat_resource)

    def on_main(self, checkpoint, resource, context,
                parameters, heat_template, **kwargs):
        neutron_client = ClientFactory.create_client("neutron", context)
        network_id = get_network_id(context)
        public_network_id = parameters.get("public_network_id")
        bank_section = checkpoint.get_resource_bank_section(network_id)

        try:
            resource_definition = bank_section.get_object("metadata")

            # Config Net
            if "network_metadata" in resource_definition:
                nets_meta = resource_definition["network_metadata"]
                self._heat_restore_networks(heat_template, nets_meta)

            # Config Subnet
            if "subnet_metadata" in resource_definition:
                subs_meta = resource_definition["subnet_metadata"]
                self._heat_restore_subnets(heat_template, nets_meta, subs_meta)

            # Config Port
            if "port_metadata" in resource_definition:
                ports_meta = resource_definition["port_metadata"]
                self._heat_restore_ports(heat_template, nets_meta,
                                         subs_meta, ports_meta)

            # Config Router
            if "router_metadata" in resource_definition:
                routers_meta = resource_definition["router_metadata"]
                self._heat_restore_routers(public_network_id,
                                           heat_template,
                                           subs_meta,
                                           routers_meta,
                                           neutron_client)

            # Config RouterInterface
            if [subs_meta is not None] and (
                    [routers_meta is not None] and [ports_meta is not None]):
                self._heat_restore_routerinterfaces(heat_template, subs_meta,
                                                    routers_meta, ports_meta)

            # Config Securiy-group
            if "security-group_metadata" in resource_definition:
                sgs_meta = resource_definition["security-group_metadata"]
                self._heat_restore_securitygroups(heat_template, sgs_meta)

        except Exception as e:
            LOG.error("restore network backup failed, network_id: %s.",
                      network_id)
            raise exception.RestoreBackupFailed(
                reason=six.text_type(e),
                resource_id=network_id,
                resource_type=constants.NETWORK_RESOURCE_TYPE
            )


class DeleteOperation(protection_plugin.Operation):
    def on_main(self, checkpoint, resource, cntxt, parameters, **kwargs):
        network_id = self._get_network_id(cntxt)
        bank_section = checkpoint.get_resource_bank_section(network_id)

        LOG.info("Deleting network backup, network_id: %s.", network_id)

        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETING)
            objects = bank_section.list_objects()
            for obj in objects:
                if obj == "status":
                    continue
                bank_section.delete_object(obj)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETED)
        except Exception as err:
            # update resource_definition backup_status
            LOG.error("Delete backup failed, network_id: %s.", network_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.DeleteBackupFailed(
                reason=err,
                resource_id=network_id,
                resource_type=self._SUPPORT_RESOURCE_TYPES)


class NeutronProtectionPlugin(protection_plugin.ProtectionPlugin):
    _SUPPORT_RESOURCE_TYPES = [constants.NETWORK_RESOURCE_TYPE]

    @classmethod
    def get_supported_resources_types(self):
        return self._SUPPORT_RESOURCE_TYPES

    @classmethod
    def get_options_schema(self, resources_type):
        return network_plugin_schemas.OPTIONS_SCHEMA

    @classmethod
    def get_restore_schema(self, resources_type):
        return network_plugin_schemas.RESTORE_SCHEMA

    @classmethod
    def get_saved_info_schema(self, resources_type):
        return network_plugin_schemas.SAVED_INFO_SCHEMA

    @classmethod
    def get_saved_info(self, metadata_store, resource):
        # TODO(chenhuayi)
        pass

    def get_protect_operation(self, resource):
        return ProtectOperation()

    def get_restore_operation(self, resource):
        return RestoreOperation()

    def get_delete_operation(self, resource):
        # TODO(chenhuayi)
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from karbor.common import constants
from karbor import exception
from karbor import resource
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protectable_plugin
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

INVALID_VOLUME_STATUS = ['error', 'deleting', 'error_deleting']


class VolumeProtectablePlugin(protectable_plugin.ProtectablePlugin):
    """Cinder volume protectable plugin"""

    _SUPPORT_RESOURCE_TYPE = constants.VOLUME_RESOURCE_TYPE

    def _client(self, context):
        self._client_instance = ClientFactory.create_client(
            "cinder",
            context)

        return self._client_instance

    def get_resource_type(self):
        return self._SUPPORT_RESOURCE_TYPE

    def get_parent_resource_types(self):
        return (constants.SERVER_RESOURCE_TYPE,
                constants.PROJECT_RESOURCE_TYPE)

    def list_resources(self, context, parameters=None):
        try:
            volumes = self._client(context).volumes.list(detailed=True)
        except Exception as e:
            LOG.exception("List all summary volumes from cinder failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(
                type=self._SUPPORT_RESOURCE_TYPE,
                id=vol.id, name=vol.name,
                extra_info={'availability_zone': vol.availability_zone})
                for vol in volumes
                if vol.status not in INVALID_VOLUME_STATUS]

    def show_resource(self, context, resource_id, parameters=None):
        try:
            volume = self._client(context).volumes.get(resource_id)
        except Exception as e:
            LOG.exception("Show a summary volume from cinder failed.")
            raise exception.ProtectableResourceNotFound(
                id=resource_id,
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            if volume.status in INVALID_VOLUME_STATUS:
                raise exception.ProtectableResourceInvalidStatus(
                    id=resource_id, type=self._SUPPORT_RESOURCE_TYPE,
                    status=volume.status)
            return resource.Resource(
                type=self._SUPPORT_RESOURCE_TYPE,
                id=volume.id, name=volume.name,
                extra_info={'availability_zone': volume.availability_zone})

    def get_dependent_resources(self, context, parent_resource):
        def _is_attached_to(vol):
            if parent_resource.type == constants.SERVER_RESOURCE_TYPE:
                return any([s.get('server_id') == parent_resource.id
                            for s in vol.attachments])
            if parent_resource.type == constants.PROJECT_RESOURCE_TYPE:
                return getattr(
                    vol,
                    'os-vol-tenant-attr:tenant_id'
                ) == parent_resource.id

        try:
            volumes = self._client(context).volumes.list(detailed=True)
        except Exception as e:
            LOG.exception("List all detailed volumes from cinder failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            return [resource.Resource(
                type=self._SUPPORT_RESOURCE_TYPE, id=vol.id, name=vol.name,
                extra_info={'availability_zone': vol.availability_zone})
                for vol in volumes if _is_attached_to(vol)]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cinderclient import exceptions as cinder_exc
import collections
from karbor.common import constants
from karbor.context import RequestContext
from karbor import exception
from karbor.resource import Resource
from karbor.services.protection import bank_plugin
from karbor.services.protection import client_factory
from karbor.services.protection.protection_plugins.volume. \
    cinder_protection_plugin import CinderBackupProtectionPlugin
from karbor.services.protection.protection_plugins.volume \
    import volume_plugin_cinder_schemas as cinder_schemas
from karbor.services.protection.restore_heat import HeatTemplate
from karbor.tests import base
from karbor.tests.unit.protection import fakes
import mock
from oslo_config import cfg
from oslo_config import fixture


ResourceNode = collections.namedtuple(
    "ResourceNode",
    ["value",
     "child_nodes"]
)

Image = collections.namedtuple(
    "Image",
    ["disk_format",
     "container_format",
     "status"]
)


def call_hooks(operation, checkpoint, resource, context, parameters, **kwargs):
    def noop(*args, **kwargs):
        pass

    hooks = (
        'on_prepare_begin',
        'on_prepare_finish',
        'on_main',
        'on_complete',
    )
    for hook_name in hooks:
        hook = getattr(operation, hook_name, noop)
        hook(checkpoint, resource, context, parameters, **kwargs)


class FakeCheckpoint(object):
    def __init__(self, section):
        super(FakeCheckpoint, self).__init__()
        self.bank_section = section
        self.id = "fake_id"

    def get_resource_bank_section(self, resource_id=None):
        return self.bank_section


class BackupResponse(object):
    def __init__(self, bkup_id, final_status, working_status, time_to_work):
        super(BackupResponse, self).__init__()
        self._final_status = final_status
        self._working_status = working_status
        self._time_to_work = time_to_work
        self._id = bkup_id

    def __call__(self, *args, **kwargs):
        res = mock.Mock()
        res.id = self._id
        if self._time_to_work > 0:
            self._time_to_work -= 1
            res.status = self._working_status
        else:
            res.status = self._final_status
        if res.status == 'not-found':
            raise cinder_exc.NotFound(403)
        return res


class RestoreResponse(object):
    def __init__(self, volume_id, raise_except=False):
        self._volume_id = volume_id
        self._raise_except = raise_except

    def __call__(self, *args, **kwargs):
        if self._raise_except:
            raise exception.KarborException()

        res = mock.Mock()
        res.volume_id = self._volume_id
        return res


class CinderProtectionPluginTest(base.TestCase):
    def setUp(self):
        super(CinderProtectionPluginTest, self).setUp()
        plugin_config = cfg.ConfigOpts()
        plugin_config_fixture = self.useFixture(fixture.Config(plugin_config))
        plugin_config_fixture.load_raw_values(
            group='cinder_backup_protection_plugin',
            poll_interval=0,
        )
        self.plugin = CinderBackupProtectionPlugin(plugin_config)
        cfg.CONF.set_default('cinder_endpoint',
                             'http://127.0.0.1:8776/v2',
                             'cinder_client')

        self.cntxt = RequestContext(user_id='demo',
                                    project_id='abcd',
                                    auth_token='efgh')
        self.cinder_client = client_factory.ClientFactory.create_client(
            "cinder", self.cntxt)

    def _get_checkpoint(self):
        fake_bank = bank_plugin.Bank(fakes.FakeBankPlugin())
        fake_bank_section = bank_plugin.BankSection(
            bank=fake_bank,
            section="fake"
        )
        return FakeCheckpoint(fake_bank_section)

    def test_get_options_schema(self):
        options_schema = self.plugin.get_options_schema(
            'OS::Cinder::Volume')
        self.assertEqual(options_schema, cinder_schemas.OPTIONS_SCHEMA)

    def test_get_restore_schema(self):
        options_schema = self.plugin.get_restore_schema(
            'OS::Cinder::Volume')
        self.assertEqual(options_schema, cinder_schemas.RESTORE_SCHEMA)

    def test_get_saved_info_schema(self):
        options_schema = self.plugin.get_saved_info_schema(
            'OS::Cinder::Volume')
        self.assertEqual(options_schema,
                         cinder_schemas.SAVED_INFO_SCHEMA)

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_protect_succeed(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        operation = self.plugin.get_protect_operation(resource)
        section.update_object = mock.MagicMock()
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            backups=mock.DEFAULT,
            volume_snapshots=mock.DEFAULT,
        ) as mocks:
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'available'
            mocks['backups'].create = BackupResponse(
                '456', 'creating', '---', 0)
            mocks['backups'].get = BackupResponse(
                '456', 'available', 'creating', 2)
            mocks['volume_snapshots'].get.return_value = BackupResponse(
                '789', 'creating', '---', 0)
            mocks['volume_snapshots'].get = BackupResponse(
                '789', 'available', 'creating', 2)
            call_hooks(operation, checkpoint, resource, self.cntxt, {})

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_protect_fail_backup(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        operation = self.plugin.get_protect_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            backups=mock.DEFAULT,
            volume_snapshots=mock.DEFAULT,
        ) as mocks:
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'available'
            mocks['backups'].backups.create = BackupResponse(
                '456', 'creating', '---', 0)
            mocks['backups'].backups.get = BackupResponse(
                '456', 'error', 'creating', 2)
            mocks['volume_snapshots'].get.return_value = BackupResponse(
                '789', 'creating', '---', 0)
            mocks['volume_snapshots'].get = BackupResponse(
                '789', 'available', 'creating', 2)
            self.assertRaises(
                exception.CreateBackupFailed,
                call_hooks,
                operation,
                checkpoint,
                resource,
                self.cntxt,
                {}
            )

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_protect_fail_snapshot(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        operation = self.plugin.get_protect_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            backups=mock.DEFAULT,
            volume_snapshots=mock.DEFAULT,
        ) as mocks:
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'available'
            mocks['backups'].backups.create = BackupResponse(
                '456', 'creating', '---', 0)
            mocks['backups'].backups.get = BackupResponse(
                '456', 'available', 'creating', 2)
            mocks['volume_snapshots'].get.return_value = BackupResponse(
                '789', 'creating', '---', 0)
            mocks['volume_snapshots'].get = BackupResponse(
                '789', 'error', 'creating', 2)
            self.assertRaises(
                exception.CreateBackupFailed,
                call_hooks,
                operation,
                checkpoint,
                resource,
                self.cntxt,
                {}
            )

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_protect_fail_volume(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        operation = self.plugin.get_protect_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            backups=mock.DEFAULT,
            volume_snapshots=mock.DEFAULT,
        ) as mocks:
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'error'
            mocks['backups'].backups.create = BackupResponse(
                '456', 'creating', '---', 0)
            mocks['backups'].backups.get = BackupResponse(
                '456', 'error', 'creating', 2)
            mocks['volume_snapshots'].get.return_value = BackupResponse(
                '789', 'creating', '---', 0)
            mocks['volume_snapshots'].get = BackupResponse(
                '789', 'available', 'creating', 2)
            self.assertRaises(
                exception.CreateBackupFailed,
                call_hooks,
                operation,
                checkpoint,
                resource,
                self.cntxt,
                {}
            )

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_delete_succeed(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        section.update_object('metadata', {
            'backup_id': '456',
        })
        operation = self.plugin.get_delete_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.object(self.cinder_client, 'backups') as backups:
            backups.delete = BackupResponse('456', 'deleting', '---', 0)
            backups.get = BackupResponse('456', 'not-found', 'deleting', 2)
            call_hooks(operation, checkpoint, resource, self.cntxt, {})

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_delete_fail(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="test",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        section.update_object('metadata', {
            'backup_id': '456',
        })
        operation = self.plugin.get_delete_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.object(self.cinder_client, 'backups') as backups:
            backups.delete = BackupResponse('456', 'deleting', '---', 0)
            backups.get = BackupResponse('456', 'error', 'deleting', 2)
            self.assertRaises(
                exception.DeleteBackupFailed,
                call_hooks,
                operation,
                checkpoint,
                resource,
                self.cntxt,
                {}
            )

    @mock.patch('karbor.services.protection.clients.cinder.create')
    @mock.patch('karbor.services.protection.protection_plugins.utils.'
                'update_resource_restore_result')
    def test_restore_succeed(self, mock_update_restore, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="fake",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        section.update_object('metadata', {
            'backup_id': '456',
        })

        parameters = {
            "restore_name": "karbor restore volume",
            "restore_description": "karbor restore",
        }

        operation = self.plugin.get_restore_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            restores=mock.DEFAULT,
        ) as mocks:
            volume_id = 456
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'available'
            mocks['restores'].restore = RestoreResponse(volume_id)
            call_hooks(operation, checkpoint, resource, self.cntxt, parameters,
                       **{'restore':  None, 'heat_template': HeatTemplate()})
            mocks['volumes'].update.assert_called_with(
                volume_id,
                **{'name': parameters['restore_name'],
                   'description': parameters['restore_description']})
            mock_update_restore.assert_called_with(
                None, resource.type, volume_id, 'available')

    @mock.patch('karbor.services.protection.clients.cinder.create')
    def test_restore_fail_volume_0(self, mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="fake",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        section.update_object('metadata', {
            'backup_id': '456',
        })

        operation = self.plugin.get_restore_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            restores=mock.DEFAULT,
        ) as mocks:
            mocks['restores'].restore = RestoreResponse(0, True)
            self.assertRaises(
                exception.KarborException, call_hooks,
                operation, checkpoint, resource, self.cntxt,
                {}, **{'restore':  None})

    @mock.patch('karbor.services.protection.clients.cinder.create')
    @mock.patch('karbor.services.protection.protection_plugins.utils.'
                'update_resource_restore_result')
    def test_restore_fail_volume_1(self, mock_update_restore,
                                   mock_cinder_create):
        resource = Resource(
            id="123",
            type=constants.VOLUME_RESOURCE_TYPE,
            name="fake",
        )
        checkpoint = self._get_checkpoint()
        section = checkpoint.get_resource_bank_section()
        section.update_object('metadata', {
            'backup_id': '456',
        })

        operation = self.plugin.get_restore_operation(resource)
        mock_cinder_create.return_value = self.cinder_client
        with mock.patch.multiple(
            self.cinder_client,
            volumes=mock.DEFAULT,
            restores=mock.DEFAULT,
        ) as mocks:
            volume_id = 456
            mocks['volumes'].get.return_value = mock.Mock()
            mocks['volumes'].get.return_value.status = 'error'
            mocks['restores'].restore = RestoreResponse(volume_id)
            self.assertRaises(
                exception.RestoreBackupFailed, call_hooks,
                operation, checkpoint, resource, self.cntxt,
                {}, **{'restore':  None})

            mock_update_restore.assert_called_with(
                None, resource.type, volume_id,
                constants.RESOURCE_STATUS_ERROR, 'Error creating volume')

    def test_get_supported_resources_types(self):
        types = self.plugin.get_supported_resources_types()
        self.assertEqual([constants.VOLUME_RESOURCE_TYPE], types)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import itertools

import karbor.api.common
import karbor.api.v1.protectables
import karbor.api.v1.providers
import karbor.common.config
import karbor.db.api
import karbor.exception
import karbor.service
import karbor.services.operationengine.engine.executors.green_thread_executor as green_thread_executor  # noqa
import karbor.services.operationengine.engine.executors.thread_pool_executor as thread_pool_executor  # noqa
import karbor.services.operationengine.engine.triggers.timetrigger.time_trigger as time_trigger  # noqa
import karbor.services.operationengine.karbor_client
import karbor.services.operationengine.manager
import karbor.services.operationengine.operations.base as base
import karbor.services.protection.clients.cinder
import karbor.services.protection.clients.glance
import karbor.services.protection.clients.heat
import karbor.services.protection.clients.manila
import karbor.services.protection.clients.neutron
import karbor.services.protection.clients.nova
import karbor.services.protection.flows.restore
import karbor.services.protection.flows.worker
import karbor.services.protection.manager
import karbor.services.protection.protection_plugins.image.image_protection_plugin as image_protection_plugin  # noqa
import karbor.services.protection.protection_plugins.share.share_snapshot_plugin as share_snapshot_plugin  # noqa
import karbor.services.protection.protection_plugins.volume.cinder_protection_plugin as cinder_protection_plugin  # noqa
import karbor.services.protection.protection_plugins.volume.volume_snapshot_plugin as volume_snapshot_plugin  # noqa
import karbor.wsgi.eventlet_server

__all__ = ['list_opts']

_opts = [
    ('clients_keystone', list(itertools.chain(
        karbor.common.config.keystone_client_opts))),
    ('operationengine', list(itertools.chain(
        green_thread_executor.green_thread_executor_opts,
        karbor.services.operationengine.manager.trigger_manager_opts))),
    ('karbor_client', list(itertools.chain(
        karbor.common.config.service_client_opts))),
    ('cinder_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.cinder.cinder_client_opts))),
    ('glance_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.glance.glance_client_opts))),
    ('heat_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.heat.heat_client_opts))),
    ('manila_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.manila.manila_client_opts))),
    ('neutron_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.neutron.neutron_client_opts))),
    ('nova_client', list(itertools.chain(
        karbor.common.config.service_client_opts,
        karbor.services.protection.clients.nova.nova_client_opts))),
    ('image_backup_plugin', list(itertools.chain(
        image_protection_plugin.image_backup_opts))),
    ('cinder_backup_protection_plugin', list(itertools.chain(
        cinder_protection_plugin.cinder_backup_opts))),
    ('volume_snapshot_plugin', list(itertools.chain(
        volume_snapshot_plugin.volume_snapshot_opts))),
    ('manila_snapshot_plugin', list(itertools.chain(
        share_snapshot_plugin.manila_snapshot_opts))),
    (None, list(itertools.chain(
        karbor.common.config.core_opts,
        karbor.common.config.debug_opts,
        karbor.common.config.global_opts,
        karbor.api.common.api_common_opts,
        karbor.api.v1.protectables.query_instance_filters_opts,
        karbor.api.v1.providers.query_provider_filters_opts,
        karbor.api.v1.providers.query_checkpoint_filters_opts,
        karbor.db.api.db_opts,
        thread_pool_executor.executor_opts,
        time_trigger.time_trigger_opts,
        base.record_operation_log_executor_opts,
        karbor.services.protection.flows.restore.sync_status_opts,
        karbor.services.protection.flows.worker.workflow_opts,
        karbor.services.protection.manager.protection_manager_opts,
        karbor.wsgi.eventlet_server.socket_opts,
        karbor.exception.exc_log_opts,
        karbor.service.service_opts)))]


def list_opts():
    return [(g, copy.deepcopy(o)) for g, o in _opts]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple
from glanceclient.v2 import images
from karbor.common import constants
from karbor.context import RequestContext
from karbor import resource
from karbor.services.protection.protectable_plugins.image import \
    ImageProtectablePlugin
from karbor.tests import base
from keystoneauth1 import session as keystone_session
import mock
from novaclient.v2 import servers
from oslo_config import cfg

CONF = cfg.CONF

image_info = namedtuple('image_info', field_names=['id', 'owner', 'name',
                                                   'status'])
server_info = namedtuple('server_info', field_names=['id', 'type', 'name',
                                                     'image'])
project_info = namedtuple('project_info', field_names=['id', 'type', 'name'])


class ImageProtectablePluginTest(base.TestCase):
    def setUp(self):
        super(ImageProtectablePluginTest, self).setUp()
        service_catalog = [{
            'type': 'image',
            'endpoints': [{'publicURL': 'http://127.0.0.1:9292'}]
        }, {
            'type': 'compute',
            'endpoints': [{'publicURL': 'http://127.0.0.1:8774/v2.1/abcd'}]
        }]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_endpoint(self, mock_generate_session):
        CONF.set_default('glance_endpoint', 'http://127.0.0.1:9292',
                         'glance_client')
        CONF.set_default('nova_endpoint', 'http://127.0.0.1:8774/v2.1',
                         'nova_client')
        plugin = ImageProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual(
            plugin._glance_client(self._context).http_client.endpoint_override,
            'http://127.0.0.1:9292')
        self.assertEqual(
            plugin._nova_client(self._context).client.endpoint_override,
            'http://127.0.0.1:8774/v2.1/abcd')

    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_create_client_by_catalog(self, mock_generate_session):
        CONF.set_default('glance_catalog_info', 'image:glance:publicURL',
                         'glance_client')
        CONF.set_default('nova_catalog_info', 'compute:nova:publicURL',
                         'nova_client')
        plugin = ImageProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        self.assertEqual(
            plugin._glance_client(self._context).http_client.endpoint_override,
            'http://127.0.0.1:9292')
        self.assertEqual(
            plugin._nova_client(self._context).client.endpoint_override,
            'http://127.0.0.1:8774/v2.1/abcd')

    def test_get_resource_type(self):
        plugin = ImageProtectablePlugin(self._context)
        self.assertEqual(
            plugin.get_resource_type(),
            constants.IMAGE_RESOURCE_TYPE)

    def test_get_parent_resource_type(self):
        plugin = ImageProtectablePlugin(self._context)
        self.assertItemsEqual(
            plugin.get_parent_resource_types(),
            (constants.SERVER_RESOURCE_TYPE, constants.PROJECT_RESOURCE_TYPE))

    @mock.patch.object(images.Controller, 'list')
    def test_list_resources(self, mokc_image_list):
        plugin = ImageProtectablePlugin(self._context)
        mokc_image_list.return_value = [
            image_info(id='123', name='name123', owner='abcd',
                       status='active'),
            image_info(id='456', name='name456', owner='efgh',
                       status='active'),
        ]
        self.assertEqual(plugin.list_resources(self._context),
                         [resource.Resource(type=constants.IMAGE_RESOURCE_TYPE,
                                            id='123', name='name123'),
                          resource.Resource(type=constants.IMAGE_RESOURCE_TYPE,
                                            id='456', name='name456')
                          ])

    @mock.patch.object(images.Controller, 'get')
    def test_show_resource(self, mock_image_get):
        image_info = namedtuple('image_info', field_names=['id', 'name',
                                                           'status'])
        plugin = ImageProtectablePlugin(self._context)
        mock_image_get.return_value = image_info(id='123', name='name123',
                                                 status='active')
        self.assertEqual(plugin.show_resource(self._context, '123'),
                         resource.Resource(type=constants.IMAGE_RESOURCE_TYPE,
                                           id='123', name='name123'))

    @mock.patch.object(images.Controller, 'get')
    @mock.patch.object(servers.ServerManager, 'get')
    @mock.patch('karbor.services.protection.client_factory.ClientFactory.'
                '_generate_session')
    def test_get_server_dependent_resources(self, mock_generate_session,
                                            mock_server_get,
                                            mock_image_get):
        vm = server_info(id='server1',
                         type=constants.SERVER_RESOURCE_TYPE,
                         name='nameserver1',
                         image=dict(id='123', name='name123'))
        image = image_info(id='123', name='name123', owner='abcd',
                           status='active')
        plugin = ImageProtectablePlugin(self._context)
        mock_generate_session.return_value = keystone_session.Session(
            auth=None)
        mock_server_get.return_value = vm
        mock_image_get.return_value = image
        self.assertEqual(plugin.get_dependent_resources(self._context, vm),
                         [resource.Resource(type=constants.IMAGE_RESOURCE_TYPE,
                                            id='123', name='name123')])

    @mock.patch.object(images.Controller, 'list')
    def test_get_project_dependent_resources(self, mock_image_list):
        project = project_info(id='abcd', type=constants.PROJECT_RESOURCE_TYPE,
                               name='nameabcd')
        plugin = ImageProtectablePlugin(self._context)
        mock_image_list.return_value = [
            image_info('123', 'abcd', 'nameabcd', 'active'),
            image_info('456', 'efgh', 'nameefgh', 'active'),
        ]
        self.assertEqual(
            plugin.get_dependent_resources(self._context, project),
            [resource.Resource(type=constants.IMAGE_RESOURCE_TYPE,
                               name='nameabcd',
                               id='123')])



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_versionedobjects import fields

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.objects import base


@base.KarborObjectRegistry.register
class ScheduledOperation(base.KarborPersistentObject, base.KarborObject,
                         base.KarborObjectDictCompat,
                         base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(),
        'name': fields.StringField(),
        'description': fields.StringField(nullable=True),
        'operation_type': fields.StringField(),
        'user_id': fields.StringField(),
        'project_id': fields.StringField(),
        'trigger_id': fields.UUIDField(),
        'operation_definition': fields.DictOfStringsField(),
        'enabled': fields.BooleanField(default=True),

        'trigger': fields.ObjectField("Trigger")
    }

    INSTANCE_OPTIONAL_JOINED_FIELDS = ['trigger']

    @staticmethod
    def _from_db_object(context, op, db_op, expected_attrs=[]):
        special_fields = set(['operation_definition'] +
                             op.INSTANCE_OPTIONAL_JOINED_FIELDS)

        normal_fields = set(op.fields) - special_fields
        for name in normal_fields:
            op[name] = db_op.get(name)

        op_definition = db_op['operation_definition']
        if op_definition:
            op['operation_definition'] = jsonutils.loads(op_definition)

        if 'trigger' in expected_attrs:
            if db_op.get('trigger', None) is None:
                op.trigger = None
            else:
                if not op.obj_attr_is_set('trigger'):
                    op.trigger = objects.Trigger(context)
                op.trigger._from_db_object(context, op.trigger,
                                           db_op['trigger'])

        op._context = context
        op.obj_reset_changes()
        return op

    @staticmethod
    def _convert_operation_definition_to_db_format(updates):
        op_definition = updates.pop('operation_definition', None)
        if op_definition is not None:
            updates['operation_definition'] = jsonutils.dumps(op_definition)

    @base.remotable_classmethod
    def get_by_id(cls, context, id, expected_attrs=[]):
        columns_to_join = [col for col in expected_attrs
                           if col in cls.INSTANCE_OPTIONAL_JOINED_FIELDS]

        db_op = db.scheduled_operation_get(context, id, columns_to_join)
        if db_op:
            return cls._from_db_object(context, cls(), db_op, expected_attrs)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))

        updates = self.karbor_obj_get_changes()
        self._convert_operation_definition_to_db_format(updates)
        db_op = db.scheduled_operation_create(self._context, updates)
        self._from_db_object(self._context, self, db_op)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        if updates and self.id:
            self._convert_operation_definition_to_db_format(updates)
            db.scheduled_operation_update(self._context,
                                          self.id,
                                          updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        if self.id:
            db.scheduled_operation_delete(self._context, self.id)


@base.KarborObjectRegistry.register
class ScheduledOperationList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('ScheduledOperation'),
    }

    @base.remotable_classmethod
    def get_by_filters(cls, context, filters, limit=None,
                       marker=None, sort_keys=None, sort_dirs=None):

        db_operation_list = db.scheduled_operation_get_all_by_filters_sort(
            context, filters, limit=limit, marker=marker,
            sort_keys=sort_keys, sort_dirs=sort_dirs)

        return base.obj_make_list(context, cls(context), ScheduledOperation,
                                  db_operation_list)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The providers api."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from webob import exc

from karbor.api import common
from karbor.api.openstack import wsgi
from karbor.common import constants
from karbor import exception
from karbor.i18n import _

from karbor import objects
import karbor.policy
from karbor.services.protection import api as protection_api
from karbor import utils

import six

query_provider_filters_opts = [
    cfg.ListOpt(
        'query_provider_filters',
        default=['name', 'description'],
        help=(
            "Provider filter options which non-admin user could use to "
            "query providers. Default values are: ['name', 'description']"
        )
    ),
]

query_checkpoint_filters_opts = [
    cfg.ListOpt(
        'query_checkpoint_filters',
        default=['project_id', 'plan_id', 'start_date', 'end_date'],
        help=(
            "Checkpoint filter options which non-admin user could use to "
            "query checkpoints. Default values are: ['project_id', "
            "'plan_id', 'start_date', 'end_date']"
        )
    ),
]

CONF = cfg.CONF
CONF.register_opts(query_provider_filters_opts)
CONF.register_opts(query_checkpoint_filters_opts)

LOG = logging.getLogger(__name__)


def check_policy(context, action):
    target = {
        'project_id': context.project_id,
        'user_id': context.user_id,
    }
    _action = 'provider:%s' % action
    karbor.policy.enforce(context, _action, target)


class ProviderViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "providers"

    def detail(self, request, provider):
        """Detailed view of a single provider."""
        provider_ref = {
            'provider': {
                'id': provider.get('id'),
                'name': provider.get('name'),
                'description': provider.get('description'),
                'extended_info_schema': provider.get('extended_info_schema'),
            }
        }
        return provider_ref

    def detail_list(self, request, providers, provider_count=None):
        """Detailed view of a list of providers."""
        return self._list_view(self.detail, request, providers,
                               provider_count,
                               self._collection_name)

    def _list_view(self, func, request, providers, provider_count,
                   coll_name=_collection_name):
        """Provide a view for a list of provider.

        :param func: Function used to format the provider data
        :param request: API request
        :param providers: List of providers in dictionary format
        :param provider_count: Length of the original list of providers
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: Provider data in dictionary format
        """
        providers_list = [func(request, provider)['provider']
                          for provider in providers]
        providers_links = self._get_collection_links(request,
                                                     providers,
                                                     coll_name,
                                                     provider_count)
        providers_dict = {
            "providers": providers_list
        }
        if providers_links:
            providers_dict['providers_links'] = providers_links

        return providers_dict


class CheckpointViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "checkpoints"

    def detail(self, request, checkpoint):
        """Detailed view of a single checkpoint."""
        checkpoint_ref = {
            'checkpoint': {
                'id': checkpoint.get('id'),
                'project_id': checkpoint.get('project_id'),
                'status': checkpoint.get('status'),
                'protection_plan': checkpoint.get('protection_plan'),
                'resource_graph': checkpoint.get('resource_graph'),
                'created_at': checkpoint.get('created_at'),
                'extra_info': checkpoint.get('extra_info'),
            }
        }
        return checkpoint_ref

    def detail_list(self, request, checkpoints, checkpoint_count=None):
        """Detailed view of a list of checkpoints."""
        return self._list_view(self.detail, request, checkpoints,
                               checkpoint_count,
                               self._collection_name)

    def _list_view(self, func, request, checkpoints, checkpoint_count,
                   coll_name=_collection_name):
        """Provide a view for a list of checkpoint.

        :param func: Function used to format the checkpoint data
        :param request: API request
        :param checkpoints: List of checkpoints in dictionary format
        :param checkpoint_count: Length of the original list of checkpoints
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: Checkpoint data in dictionary format
        """
        checkpoints_list = [func(request, checkpoint)['checkpoint']
                            for checkpoint in checkpoints]
        checkpoints_links = self._get_collection_links(request,
                                                       checkpoints,
                                                       coll_name,
                                                       checkpoint_count)
        checkpoints_dict = {
            "checkpoints": checkpoints_list
        }
        if checkpoints_links:
            checkpoints_dict['checkpoints_links'] = checkpoints_links

        return checkpoints_dict


class ProvidersController(wsgi.Controller):
    """The Providers API controller for the OpenStack API."""

    _view_builder_class = ProviderViewBuilder

    def __init__(self):
        self.protection_api = protection_api.API()
        self._checkpoint_view_builder = CheckpointViewBuilder()
        super(ProvidersController, self).__init__()

    def show(self, req, id):
        """Return data about the given provider id."""
        context = req.environ['karbor.context']

        LOG.info("Show provider with id: %s", id)

        try:
            provider = self._provider_get(context, id)
        except exception.ProviderNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        LOG.info("Show provider request issued successfully.",
                 resource={'id': provider.get("id")})
        return self._view_builder.detail(req, provider)

    def index(self, req):
        """Returns a list of providers, transformed through view builder."""
        context = req.environ['karbor.context']

        LOG.info("Show provider list", context=context)

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_provider_filter_options())

        utils.check_filters(filters)
        providers = self._get_all(context, marker, limit,
                                  sort_keys=sort_keys,
                                  sort_dirs=sort_dirs,
                                  filters=filters,
                                  offset=offset)

        retval_providers = self._view_builder.detail_list(req, providers)

        LOG.info("Show provider list request issued successfully.")

        return retval_providers

    def _get_all(self, context, marker=None, limit=None, sort_keys=None,
                 sort_dirs=None, filters=None, offset=None):
        check_policy(context, 'get_all')

        if filters is None:
            filters = {}

        try:
            if limit is not None:
                limit = int(limit)
                if limit <= 0:
                    msg = _('limit param must be positive')
                    raise exception.InvalidInput(reason=msg)
        except ValueError:
            msg = _('limit param must be an integer')
            raise exception.InvalidInput(reason=msg)

        if filters:
            LOG.debug("Searching by: %s.", six.text_type(filters))

        providers = self.protection_api.list_providers(
            context, marker, limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters,
            offset=offset)

        LOG.info("Get all providers completed successfully.")
        return providers

    def _get_provider_filter_options(self):
        """Return provider search options allowed by non-admin."""
        return CONF.query_provider_filters

    def _get_checkpoint_filter_options(self):
        """Return checkpoint search options allowed by non-admin."""
        return CONF.query_checkpoint_filters

    def _provider_get(self, context, provider_id):
        if not uuidutils.is_uuid_like(provider_id):
            msg = _("Invalid provider id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            check_policy(context, 'get')
        except exception.PolicyNotAuthorized:
            # raise ProviderNotFound instead to make sure karbor behaves
            # as it used to
            raise exception.ProviderNotFound(provider_id=provider_id)

        provider = self.protection_api.show_provider(context, provider_id)

        LOG.info("Provider info retrieved successfully.")
        return provider

    def checkpoints_index(self, req, provider_id):
        """Returns a list of checkpoints, transformed through view builder."""
        context = req.environ['karbor.context']

        LOG.info("Show checkpoints list. provider_id:%s", provider_id)

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params

        utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_checkpoint_filter_options())

        utils.check_filters(filters)
        checkpoints = self._checkpoints_get_all(
            context, provider_id, marker, limit,
            sort_keys=sort_keys, sort_dirs=sort_dirs,
            filters=filters, offset=offset)

        retval_checkpoints = self._checkpoint_view_builder.detail_list(
            req, checkpoints)

        LOG.info("Show checkpoints list request issued successfully.")
        return retval_checkpoints

    def _checkpoints_get_all(self, context, provider_id, marker=None,
                             limit=None, sort_keys=None, sort_dirs=None,
                             filters=None, offset=None):
        check_policy(context, 'checkpoint_get_all')

        if filters is None:
            filters = {}

        try:
            if limit is not None:
                limit = int(limit)
                if limit <= 0:
                    msg = _('limit param must be positive')
                    raise exception.InvalidInput(reason=msg)
        except ValueError:
            msg = _('limit param must be an integer')
            raise exception.InvalidInput(reason=msg)

        if filters:
            LOG.debug("Searching by: %s.", six.text_type(filters))

        checkpoints = self.protection_api.list_checkpoints(
            context, provider_id, marker, limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters,
            offset=offset)

        LOG.info("Get all checkpoints completed successfully.")
        return checkpoints

    def checkpoints_create(self, req, provider_id, body):
        """Creates a new checkpoint."""
        if not self.is_valid_body(body, 'checkpoint'):
            raise exc.HTTPUnprocessableEntity()

        context = req.environ['karbor.context']

        LOG.debug('Create checkpoint request '
                  'body: %s provider_id:%s', body, provider_id)

        check_policy(context, 'checkpoint_create')
        checkpoint = body['checkpoint']
        LOG.debug('Create checkpoint request checkpoint: %s',
                  checkpoint)

        if not provider_id:
            msg = _("provider_id must be provided when creating "
                    "a checkpoint.")
            raise exception.InvalidInput(reason=msg)

        plan_id = checkpoint.get("plan_id")

        if not plan_id:
            msg = _("plan_id must be provided when creating "
                    "a checkpoint.")
            raise exception.InvalidInput(reason=msg)

        if not uuidutils.is_uuid_like(plan_id):
            msg = _("Invalid plan id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        plan = objects.Plan.get_by_id(context, plan_id)
        if not plan:
            raise exception.PlanNotFound(plan_id=plan_id)

        # check the provider_id
        if provider_id != plan.get("provider_id"):
            msg = _("The parameter provider_id is not the same as "
                    "the value in the plan.")
            raise exception.InvalidPlan(reason=msg)

        extra_info = checkpoint.get("extra_info", None)
        if extra_info is not None:
            if not isinstance(extra_info, dict):
                msg = _("The extra_info in checkpoint must be a dict when "
                        "creating a checkpoint.")
                raise exception.InvalidInput(reason=msg)
            elif not all(map(lambda s: isinstance(s, six.string_types),
                             extra_info.keys())):
                msg = _("Key of extra_info in checkpoint must be string when"
                        "creating a checkpoint.")
                raise exception.InvalidInput(reason=msg)

        checkpoint_extra_info = None
        if extra_info is not None:
            checkpoint_extra_info = jsonutils.dumps(extra_info)
        checkpoint_properties = {
            'project_id': context.project_id,
            'status': constants.CHECKPOINT_STATUS_PROTECTING,
            'provider_id': provider_id,
            "protection_plan": {
                "id": plan.get("id"),
                "name": plan.get("name"),
                "resources": plan.get("resources"),
            },
            "extra_info": checkpoint_extra_info
        }
        try:
            checkpoint_id = self.protection_api.protect(context, plan,
                                                        checkpoint_properties)
        except Exception as error:
            msg = _("Create checkpoint failed: %s") % error
            raise exc.HTTPBadRequest(explanation=msg)

        checkpoint_properties['id'] = checkpoint_id

        LOG.info("Create the checkpoint successfully. checkpoint_id:%s",
                 checkpoint_id)
        returnval = self._checkpoint_view_builder.detail(
            req, checkpoint_properties)
        return returnval

    def checkpoints_show(self, req, provider_id, checkpoint_id):
        """Return data about the given checkpoint id."""
        context = req.environ['karbor.context']

        LOG.info("Show checkpoint with id: %s.", checkpoint_id)
        LOG.info("provider_id: %s.", provider_id)

        try:
            checkpoint = self._checkpoint_get(context, provider_id,
                                              checkpoint_id)
        except exception.CheckpointNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        LOG.info("Show checkpoint request issued successfully.")
        LOG.info("checkpoint: %s", checkpoint)
        retval = self._checkpoint_view_builder.detail(req, checkpoint)
        LOG.info("retval: %s", retval)
        return retval

    def _checkpoint_get(self, context, provider_id, checkpoint_id):
        if not uuidutils.is_uuid_like(provider_id):
            msg = _("Invalid provider id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        if not uuidutils.is_uuid_like(checkpoint_id):
            msg = _("Invalid checkpoint id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        try:
            check_policy(context, 'checkpoint_get')
        except exception.PolicyNotAuthorized:
            # raise CheckpointNotFound instead to make sure karbor behaves
            # as it used to
            raise exception.CheckpointNotFound(checkpoint_id=checkpoint_id)

        checkpoint = self.protection_api.show_checkpoint(
            context, provider_id, checkpoint_id)

        if checkpoint is None:
            raise exception.CheckpointNotFound(checkpoint_id=checkpoint_id)

        LOG.info("Checkpoint info retrieved successfully.")
        return checkpoint

    def checkpoints_delete(self, req, provider_id, checkpoint_id):
        """Delete a checkpoint."""
        context = req.environ['karbor.context']

        LOG.info("Delete checkpoint with id: %s.", checkpoint_id)
        LOG.info("provider_id: %s.", provider_id)

        if not uuidutils.is_uuid_like(provider_id):
            msg = _("Invalid provider id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        if not uuidutils.is_uuid_like(checkpoint_id):
            msg = _("Invalid checkpoint id provided.")
            raise exc.HTTPBadRequest(explanation=msg)

        check_policy(context, 'checkpoint_delete')
        self.protection_api.delete(context,
                                   provider_id,
                                   checkpoint_id)

        LOG.info("Delete checkpoint request issued successfully.")
        return {}


def create_resource():
    return wsgi.Resource(ProvidersController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import eventlet


def wait_until_true(predicate, timeout=60, sleep=1, exception=None):
    """Wait until callable predicate is evaluated as True

    :param predicate: Callable deciding whether waiting should continue.
    Best practice is to instantiate predicate with functools.partial()
    :param timeout: Timeout in seconds how long should function wait.
    :param sleep: Polling interval for results in seconds.
    :param exception: Exception class for eventlet.Timeout.
    (see doc for eventlet.Timeout for more information)

    """
    with eventlet.timeout.Timeout(timeout, exception):
        while not predicate():
            eventlet.sleep(sleep)


def wait_until_is_and_return(predicate, timeout=5, sleep=1, exception=None):
    container = {}

    def internal_predicate():
        container['value'] = predicate()
        return container['value']

    wait_until_true(internal_predicate, timeout, sleep, exception)
    return container.get('value')


def wait_until_none(predicate, timeout=5, sleep=1, exception=None):
    def internal_predicate():
        ret = predicate()
        if ret:
            return False
        return True
    wait_until_true(internal_predicate, timeout, sleep, exception)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from abc import ABCMeta
from abc import abstractmethod
import six

from karbor import loadables


@six.add_metaclass(ABCMeta)
class BaseTrigger(object):
    """Trigger base class that all Triggers should inherit from"""

    TRIGGER_TYPE = ""

    def __init__(self, trigger_id, trigger_property, executor):
        super(BaseTrigger, self).__init__()

        self._id = trigger_id
        self._operation_ids = set()
        self._executor = executor

    @abstractmethod
    def shutdown(self):
        pass

    @abstractmethod
    def register_operation(self, operation_id, **kwargs):
        pass

    @abstractmethod
    def unregister_operation(self, operation_id, **kwargs):
        pass

    @abstractmethod
    def update_trigger_property(self, trigger_property):
        pass

    @classmethod
    @abstractmethod
    def check_trigger_definition(cls, trigger_definition):
        pass

    @classmethod
    @abstractmethod
    def check_configuration(cls):
        pass

    def has_operations(self):
        return (len(self._operation_ids) != 0)


class TriggerHandler(loadables.BaseLoader):

    def __init__(self):
        super(TriggerHandler, self).__init__(BaseTrigger)


def all_triggers():
    """Get all trigger classes."""
    return TriggerHandler().get_all_classes()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from oslo_config import cfg
import oslo_messaging

from karbor import exception
from karbor.resource import Resource
from karbor.services.protection.flows import worker as flow_manager
from karbor.services.protection import manager
from karbor.services.protection import protectable_registry
from karbor.services.protection import provider

from karbor.tests import base
from karbor.tests.unit.protection import fakes

CONF = cfg.CONF


class ProtectionServiceTest(base.TestCase):
    def setUp(self):
        self.load_engine = flow_manager.Worker._load_engine
        flow_manager.Worker._load_engine = mock.Mock()
        flow_manager.Worker._load_engine.return_value = fakes.FakeFlowEngine()
        super(ProtectionServiceTest, self).setUp()
        self.pro_manager = manager.ProtectionManager()
        self.protection_plan = fakes.fake_protection_plan()

    @mock.patch.object(protectable_registry.ProtectableRegistry,
                       'list_resource_types')
    def test_list_protectable_types(self, mocker):
        expected = ["OS::Nova::Server",
                    "OS::Cinder::Volume"]
        mocker.return_value = expected
        result = self.pro_manager.list_protectable_types(None)
        self.assertEqual(expected, result)

    def test_show_protectable_type(self):
        def mock_plugins(self):
            self._plugin_map = {
                "OS::Nova::Server": server_plugin,
                "OS::Cinder::Volume": volume_plugin
            }

        server_plugin = fakes.FakeProtectablePlugin()
        server_plugin.get_resource_type = mock.MagicMock(
            return_value="OS::Nova::Server")
        volume_plugin = fakes.FakeProtectablePlugin()
        volume_plugin.get_parent_resource_types = mock.MagicMock(
            return_value=["OS::Nova::Server"])

        protectable_registry.ProtectableRegistry.load_plugins = mock_plugins

        result = self.pro_manager.show_protectable_type(None,
                                                        "OS::Nova::Server")
        self.assertEqual("OS::Nova::Server", result["name"])
        self.assertEqual({"OS::Cinder::Volume", "OS::Glance::Image",
                          "OS::Neutron::Network"},
                         set(result["dependent_types"]))

    @mock.patch.object(protectable_registry.ProtectableRegistry,
                       'show_resource')
    def test_show_protectable_instance(self, mocker):
        mocker.return_value = Resource(type='OS::Nova::Server',
                                       id='123456',
                                       name='name123')
        fake_cntx = mock.MagicMock()

        result = self.pro_manager.show_protectable_instance(
            fake_cntx, 'OS::Nova::Server', '123456')
        self.assertEqual(
            {'id': '123456', 'name': 'name123', 'type': 'OS::Nova::Server',
             'extra_info': None},
            result)

    @mock.patch.object(protectable_registry.ProtectableRegistry,
                       'list_resources')
    def test_list_protectable_instances(self, mocker):
        mocker.return_value = [Resource(type='OS::Nova::Server',
                                        id='123456',
                                        name='name123'),
                               Resource(type='OS::Nova::Server',
                                        id='654321',
                                        name='name654')]
        fake_cntx = mock.MagicMock()

        result = self.pro_manager.list_protectable_instances(
            fake_cntx, 'OS::Nova::Server')
        self.assertEqual([{'id': '123456', 'name': 'name123',
                           'extra_info': None},
                          {'id': '654321', 'name': 'name654',
                           'extra_info': None}],
                         result)

    @mock.patch.object(protectable_registry.ProtectableRegistry,
                       'fetch_dependent_resources')
    def test_list_protectable_dependents(self, mocker):
        mocker.return_value = [Resource(type='OS::Cinder::Volume',
                                        id='123456', name='name123'),
                               Resource(type='OS::Cinder::Volume',
                                        id='654321', name='name654')]
        fake_cntx = mock.MagicMock()

        result = self.pro_manager.list_protectable_dependents(
            fake_cntx, 'fake_id', 'OS::Nova::Server')
        self.assertEqual([{'type': 'OS::Cinder::Volume', 'id': '123456',
                           'name': 'name123', 'extra_info': None},
                          {'type': 'OS::Cinder::Volume', 'id': '654321',
                           'name': 'name654', 'extra_info': None}],
                         result)

    @mock.patch.object(provider.ProviderRegistry, 'show_provider')
    def test_protect(self, mock_provider):
        mock_provider.return_value = fakes.FakeProvider()
        self.pro_manager.protect(None, fakes.fake_protection_plan())

    @mock.patch.object(flow_manager.Worker, 'get_flow')
    def test_protect_in_error(self, mock_flow):
        mock_flow.side_effect = Exception()
        self.assertRaises(oslo_messaging.ExpectedException,
                          self.pro_manager.protect,
                          None,
                          fakes.fake_protection_plan())

    @mock.patch.object(provider.ProviderRegistry, 'show_provider')
    def test_show_checkpoint(self, mock_provider):
        mock_provider.return_value = fakes.FakeProvider()
        context = mock.MagicMock()
        cp = self.pro_manager.show_checkpoint(context, 'provider1',
                                              'fake_checkpoint')
        self.assertEqual('fake_checkpoint', cp['id'])

    @mock.patch.object(provider.ProviderRegistry, 'show_provider')
    @mock.patch.object(fakes.FakeCheckpointCollection, 'get')
    def test_show_checkpoint_not_found(self, mock_provider,
                                       mock_cp_collection_get):
        mock_provider.return_value = fakes.FakeProvider()
        context = mock.MagicMock()
        mock_cp_collection_get.side_effect = exception.CheckpointNotFound(
            checkpoint_id='123')
        self.assertRaises(oslo_messaging.ExpectedException,
                          self.pro_manager.show_checkpoint,
                          context,
                          'provider1',
                          'non_existent_checkpoint')

    def tearDown(self):
        flow_manager.Worker._load_engine = self.load_engine
        super(ProtectionServiceTest, self).tearDown()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The protectables api."""
from oslo_config import cfg
from oslo_log import log as logging

from webob import exc

from karbor.api import common
from karbor.api.openstack import wsgi
from karbor import exception
from karbor.i18n import _

import karbor.policy
from karbor.services.protection import api as protection_api
from karbor import utils

import six

query_instance_filters_opts = [
    cfg.ListOpt(
        'query_instance_filters',
        default=['status'],
        help=(
            "Instance filter options which non-admin user could use to "
            "query instances. Default values are: ['status']"
        )
    ),
]
CONF = cfg.CONF
CONF.register_opts(query_instance_filters_opts)
LOG = logging.getLogger(__name__)


def check_policy(context, action):
    target = {
        'project_id': context.project_id,
        'user_id': context.user_id,
    }
    _action = 'protectable:%s' % action
    karbor.policy.enforce(context, _action, target)


class ProtectableViewBuilder(common.ViewBuilder):
    """Model a server API response as a python dictionary."""

    _collection_name = "protectables"

    def show(self, request, protectable_type):
        """Detailed view of a single protectable_type."""
        protectable_type_ref = {
            'protectable_type': {
                'name': protectable_type.get('name'),
                'dependent_types': protectable_type.get('dependent_types'),
            }
        }
        return protectable_type_ref

    def detail(self, request, instance):
        """Detailed view of a single instance."""
        instance_ref = {
            'instance': {
                'id': instance.get('id'),
                'type': instance.get('type'),
                'name': instance.get('name'),
                'extra_info': instance.get('extra_info'),
                'dependent_resources': instance.get('dependent_resources'),
            }
        }
        return instance_ref

    def detail_list(self, request, instances, instance_count=None):
        """Detailed view of a list of instances."""
        return self._list_view(self.detail, request, instances,
                               instance_count,
                               'instances')

    def _list_view(self, func, request, instances, instance_count,
                   coll_name=_collection_name):
        """Provide a view for a list of instance.

        :param func: Function used to format the instance data
        :param request: API request
        :param instances: List of instances in dictionary format
        :param instance_count: Length of the original list of instances
        :param coll_name: Name of collection, used to generate the next link
                          for a pagination query
        :returns: instance data in dictionary format
        """
        instances_list = [func(request, instance)['instance']
                          for instance in instances]
        instances_links = self._get_collection_links(request,
                                                     instances,
                                                     coll_name,
                                                     instance_count)
        instances_dict = {
            "instances": instances_list
        }
        if instances_links:
            instances_dict['instances_links'] = instances_links

        return instances_dict


class ProtectablesController(wsgi.Controller):
    """The Protectables API controller for the OpenStack API."""

    _view_builder_class = ProtectableViewBuilder

    def __init__(self):
        self.protection_api = protection_api.API()
        super(ProtectablesController, self).__init__()

    def show(self, req, id):
        """Return data about the given protectable_type."""
        context = req.environ['karbor.context']
        protectable_type = id

        LOG.info("Show the information of a given protectable type: %s",
                 protectable_type)

        protectable_types = self._get_all(context)

        if protectable_type not in protectable_types:
            msg = _("Invalid protectable type provided.")
            raise exception.InvalidInput(reason=msg)

        check_policy(context, 'get')
        try:
            retval_protectable_type = self.protection_api.\
                show_protectable_type(context, protectable_type)
        except exception.ProtectableTypeNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)

        LOG.info("Show the protectable type information issued successfully.")
        return self._view_builder.show(req, retval_protectable_type)

    def index(self, req):
        """Returns a list of protectable_types,

        transformed through view builder.
        """
        context = req.environ['karbor.context']
        LOG.info("Show protectable type list", context=context)

        protectable_types = self._get_all(context)
        retval_protectable_types = {
            "protectable_type": protectable_types
        }

        LOG.info("Show protectable type list request issued successfully.")
        return retval_protectable_types

    def _get_all(self, context):
        check_policy(context, 'get_all')

        protectable_types = self.protection_api.list_protectable_types(context)

        LOG.info("Get all protectable types completed successfully.")
        return protectable_types

    def instances_index(self, req, protectable_type):
        """Return data about the given protectable_type."""
        context = req.environ['karbor.context']
        LOG.info("Show the instances of a given protectable type: %s",
                 protectable_type)

        params = req.params.copy()
        marker, limit, offset = common.get_pagination_params(params)
        sort_keys, sort_dirs = common.get_sort_params(params)
        filters = params
        utils.check_filters(filters)
        parameters = filters.get("parameters", None)

        if parameters is not None:
            if not isinstance(parameters, dict):
                msg = _("The parameters must be a dict.")
                raise exception.InvalidInput(reason=msg)

        utils.remove_invalid_filter_options(
            context,
            filters,
            self._get_instance_filter_options())

        protectable_types = self._get_all(context)

        if protectable_type not in protectable_types:
            msg = _("Invalid protectable type provided.")
            raise exception.InvalidInput(reason=msg)

        instances = self._instances_get_all(
            context, protectable_type, marker, limit,
            sort_keys=sort_keys, sort_dirs=sort_dirs,
            filters=filters, offset=offset, parameters=parameters)

        for instance in instances:
            protectable_id = instance.get("id")
            instance["type"] = protectable_type
            if protectable_id is None:
                raise exception.InvalidProtectableInstance()
            dependents = self.protection_api.list_protectable_dependents(
                context, protectable_id, protectable_type)
            instance["dependent_resources"] = dependents

        retval_instances = self._view_builder.detail_list(req, instances)

        return retval_instances

    def _instances_get_all(self, context, protectable_type, marker=None,
                           limit=None, sort_keys=None, sort_dirs=None,
                           filters=None, offset=None, parameters=None):
        check_policy(context, 'get_all')

        if filters is None:
            filters = {}

        try:
            if limit is not None:
                limit = int(limit)
                if limit <= 0:
                    msg = _('limit param must be positive')
                    raise exception.InvalidInput(reason=msg)
        except ValueError:
            msg = _('limit param must be an integer')
            raise exception.InvalidInput(reason=msg)

        if filters:
            LOG.debug("Searching by: %s.", six.text_type(filters))

        instances = self.protection_api.list_protectable_instances(
            context, protectable_type, marker, limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            filters=filters,
            offset=offset,
            parameters=parameters)

        LOG.info("Get all instances completed successfully.")
        return instances

    def _get_instance_filter_options(self):
        """Return instance search options allowed by non-admin."""
        return CONF.query_instance_filters

    def instances_show(self, req, protectable_type, protectable_id):
        """Return a instance about the given protectable_type and id."""

        context = req.environ['karbor.context']
        params = req.params.copy()
        utils.check_filters(params)
        parameters = params.get("parameters", None)

        LOG.info("Show the instance of a given protectable type: %s",
                 protectable_type)

        if parameters is not None:
            if not isinstance(parameters, dict):
                msg = _("The parameters must be a dict.")
                raise exception.InvalidInput(reason=msg)

        protectable_types = self._get_all(context)

        if protectable_type not in protectable_types:
            msg = _("Invalid protectable type provided.")
            raise exception.InvalidInput(reason=msg)

        try:
            instance = self.protection_api.show_protectable_instance(
                context, protectable_type, protectable_id,
                parameters=parameters)
        except exception.ProtectableResourceNotFound as error:
            raise exc.HTTPNotFound(explanation=error.msg)
        except Exception as err:
            raise exc.HTTPInternalServerError(
                explanation=six.text_type(err))

        if instance is None:
            msg = _("The instance doesn't exist.")
            raise exc.HTTPInternalServerError(explanation=msg)

        dependents = self.protection_api.list_protectable_dependents(
            context, protectable_id, protectable_type)
        instance["dependent_resources"] = dependents

        retval_instance = self._view_builder.detail(req, instance)
        return retval_instance


def create_resource():
    return wsgi.Resource(ProtectablesController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_config import cfg

from karbor.context import RequestContext
from karbor.services.protection.clients import cinder
from karbor.tests import base


class CinderClientTest(base.TestCase):
    def setUp(self):
        super(CinderClientTest, self).setUp()

        self._public_url = 'http://127.0.0.1:8776/v3/abcd'

        service_catalog = [
            {'type': 'volumev3',
             'name': 'cinderv3',
             'endpoints': [{'publicURL': self._public_url}],
             },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    def test_create_client_by_endpoint(self):
        cfg.CONF.set_default('cinder_endpoint',
                             'http://127.0.0.1:8776/v3',
                             'cinder_client')
        client = cinder.create(self._context, cfg.CONF)
        self.assertEqual('volumev3', client.client.service_type)
        self.assertEqual('http://127.0.0.1:8776/v3/abcd',
                         client.client.management_url)

    def test_create_client_by_catalog(self):
        client = cinder.create(self._context, cfg.CONF)
        self.assertEqual('volumev3', client.client.service_type)
        self.assertEqual('http://127.0.0.1:8776/v3/abcd',
                         client.client.management_url)

    @mock.patch('karbor.services.protection.clients.utils.get_url')
    @mock.patch('cinderclient.client.Client')
    def test_create_client(self, create, get_url):
        get_url.return_value = self._public_url

        client_config = cfg.CONF[cinder.CONFIG_GROUP]
        client_version = cinder.CINDERCLIENT_VERSION
        session = object()
        args = {
            'project_id': self._context.project_id,
            'cacert': client_config.cinder_ca_cert_file,
            'insecure': client_config.cinder_auth_insecure,
        }

        cinder.create(self._context, cfg.CONF)
        create.assert_called_with(client_version, **args)

        cinder.create(self._context, cfg.CONF, session=session)
        create.assert_called_with(client_version,
                                  endpoint_override=self._public_url,
                                  session=session)
